# Copyright (C) 2015 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import argparse
import gl_XML

class exec_info():
    """Information relating GL APIs to a function.

    Each of the four attributes of this class, compatibility, core, es1, and
    es2, specify the minimum API version where a function can possibly exist
    in Mesa.  The version is specified as an integer of (real GL version *
    10).  For example, glCreateProgram was added in OpenGL 2.0, so
    compatibility=20 and core=31.

    If the attribute is None, then it cannot be supported by that
    API.  For example, glNewList was removed from core profiles, so
    compatibility=10 and core=None.

    Each of the attributes that is not None must have a valid value.  The
    valid ranges are:

        compatibility: [10, )
        core: [31, )
        es1: [10, 11]
        es2: [20, )

    These ranges are enforced by the constructor.
    """
    def __init__(self, compatibility=None, core=None, es1=None, es2=None):
        if compatibility is not None:
            assert isinstance(compatibility, int)
            assert compatibility >= 10

        if core is not None:
            assert isinstance(core, int)
            assert core >= 31

        if es1 is not None:
            assert isinstance(es1, int)
            assert es1 == 10 or es1 == 11

        if es2 is not None:
            assert isinstance(es2, int)
            assert es2 >= 20

        self.compatibility = compatibility
        self.core = core
        self.es1 = es1
        self.es2 = es2

functions = {
    # OpenGL 3.1 / GL_ARB_texture_buffer_object.
    "TexBuffer": exec_info(compatibility=20, core=31, es2=31),

    # OpenGL 3.2 / GL_OES_geometry_shader.
    "FramebufferTexture": exec_info(compatibility=32, core=32, es2=31),

    # OpenGL 4.0 / GL_ARB_shader_subroutines. Mesa only exposes this
    # extension with core profile.
    "GetSubroutineUniformLocation": exec_info(compatibility=31, core=31),
    "GetSubroutineIndex": exec_info(compatibility=31, core=31),
    "GetActiveSubroutineUniformiv": exec_info(compatibility=31, core=31),
    "GetActiveSubroutineUniformName": exec_info(compatibility=31, core=31),
    "GetActiveSubroutineName": exec_info(compatibility=31, core=31),
    "UniformSubroutinesuiv": exec_info(compatibility=31, core=31),
    "GetUniformSubroutineuiv": exec_info(compatibility=31, core=31),
    "GetProgramStageiv": exec_info(compatibility=31, core=31),

    # OpenGL 4.0 / GL_ARB_gpu_shader_fp64.  The extension spec says:
    #
    #     "OpenGL 3.2 and GLSL 1.50 are required."
    "Uniform1d": exec_info(compatibility=32, core=32),
    "Uniform2d": exec_info(compatibility=32, core=32),
    "Uniform3d": exec_info(compatibility=32, core=32),
    "Uniform4d": exec_info(compatibility=32, core=32),
    "Uniform1dv": exec_info(compatibility=32, core=32),
    "Uniform2dv": exec_info(compatibility=32, core=32),
    "Uniform3dv": exec_info(compatibility=32, core=32),
    "Uniform4dv": exec_info(compatibility=32, core=32),
    "UniformMatrix2dv": exec_info(compatibility=32, core=32),
    "UniformMatrix3dv": exec_info(compatibility=32, core=32),
    "UniformMatrix4dv": exec_info(compatibility=32, core=32),
    "UniformMatrix2x3dv": exec_info(compatibility=32,core=32),
    "UniformMatrix2x4dv": exec_info(compatibility=32, core=32),
    "UniformMatrix3x2dv": exec_info(compatibility=32, core=32),
    "UniformMatrix3x4dv": exec_info(compatibility=32, core=32),
    "UniformMatrix4x2dv": exec_info(compatibility=32, core=32),
    "UniformMatrix4x3dv": exec_info(compatibility=32, core=32),
    "GetUniformdv": exec_info(compatibility=32, core=32),

    # OpenGL 4.1 / GL_ARB_vertex_attrib_64bit.  The extension spec says:
    #
    #     "OpenGL 3.0 and GLSL 1.30 are required.
    #
    #     ARB_gpu_shader_fp64 (or equivalent functionality) is required."
    #
    # For Mesa this effectively means OpenGL 3.2 is required.  It seems
    # unlikely that Mesa will ever get support for any of the NV extensions
    # that add "equivalent functionality."
    "VertexAttribL1d": exec_info(compatibility=32, core=32),
    "VertexAttribL2d": exec_info(compatibility=32, core=32),
    "VertexAttribL3d": exec_info(compatibility=32, core=32),
    "VertexAttribL4d": exec_info(compatibility=32, core=32),
    "VertexAttribL1dv": exec_info(compatibility=32, core=32),
    "VertexAttribL2dv": exec_info(compatibility=32, core=32),
    "VertexAttribL3dv": exec_info(compatibility=32, core=32),
    "VertexAttribL4dv": exec_info(compatibility=32, core=32),
    "VertexAttribLPointer": exec_info(compatibility=32, core=32),
    "GetVertexAttribLdv": exec_info(compatibility=32, core=32),

    # OpenGL 4.1 / GL_ARB_viewport_array.  The extension spec says:
    #
    #     "OpenGL 3.2 or the EXT_geometry_shader4 or ARB_geometry_shader4
    #     extensions are required."
    #
    # Mesa does not support either of the geometry shader extensions, so
    # OpenGL 3.2 is required.
    "ViewportArrayv": exec_info(compatibility=32, core=32, es2=31),
    "ViewportIndexedf": exec_info(compatibility=32, core=32, es2=31),
    "ViewportIndexedfv": exec_info(compatibility=32, core=32, es2=31),
    "ScissorArrayv": exec_info(compatibility=32, core=32, es2=31),
    "ScissorIndexed": exec_info(compatibility=32, core=32, es2=31),
    "ScissorIndexedv": exec_info(compatibility=32, core=32, es2=31),
    "DepthRangeArrayv": exec_info(compatibility=32, core=32),
    "DepthRangeIndexed": exec_info(compatibility=32, core=32),
    # GetFloati_v also GL_ARB_shader_atomic_counters
    # GetDoublei_v also GL_ARB_shader_atomic_counters

    # OpenGL 4.3 / GL_ARB_texture_buffer_range.
    "TexBufferRange": exec_info(compatibility=20, core=31, es2=31),

    # OpenGL 4.3 / GL_ARB_framebuffer_no_attachments.  Mesa can expose the
    # extension with OpenGL 3.0.
    "FramebufferParameteri": exec_info(compatibility=30, core=31, es2=31),
    "GetFramebufferParameteriv": exec_info(compatibility=30, core=31, es2=31),

    # OpenGL 4.3 / GL_MESA_framebuffer_flip_y.
    "FramebufferParameteriMESA": exec_info(core=31, es2=30),
    "GetFramebufferParameterivMESA": exec_info(core=31, es2=30),

    # OpenGL 4.5 / GL_ARB_direct_state_access.   Mesa can expose the extension
    # with core profile.
    "CreateTransformFeedbacks": exec_info(compatibility=31, core=31),
    "TransformFeedbackBufferBase": exec_info(compatibility=31, core=31),
    "TransformFeedbackBufferRange": exec_info(compatibility=31, core=31),
    "GetTransformFeedbackiv": exec_info(compatibility=31, core=31),
    "GetTransformFeedbacki_v": exec_info(compatibility=31, core=31),
    "GetTransformFeedbacki64_v": exec_info(compatibility=31, core=31),
    "CreateBuffers": exec_info(compatibility=31, core=31),
    "NamedBufferStorage": exec_info(compatibility=31, core=31),
    "NamedBufferData": exec_info(compatibility=31, core=31),
    "NamedBufferSubData": exec_info(compatibility=31, core=31),
    "CopyNamedBufferSubData": exec_info(compatibility=31, core=31),
    "ClearNamedBufferData": exec_info(compatibility=31, core=31),
    "ClearNamedBufferSubData": exec_info(compatibility=31, core=31),
    "MapNamedBuffer": exec_info(compatibility=31, core=31),
    "MapNamedBufferRange": exec_info(compatibility=31, core=31),
    "UnmapNamedBuffer": exec_info(compatibility=31, core=31),
    "FlushMappedNamedBufferRange": exec_info(compatibility=31, core=31),
    "GetNamedBufferParameteriv": exec_info(compatibility=31, core=31),
    "GetNamedBufferParameteri64v": exec_info(compatibility=31, core=31),
    "GetNamedBufferPointerv": exec_info(compatibility=31, core=31),
    "GetNamedBufferSubData": exec_info(compatibility=31, core=31),
    "CreateFramebuffers": exec_info(compatibility=31, core=31),
    "NamedFramebufferRenderbuffer": exec_info(compatibility=31, core=31),
    "NamedFramebufferParameteri": exec_info(compatibility=31, core=31),
    "NamedFramebufferTexture": exec_info(compatibility=31, core=31),
    "NamedFramebufferTextureLayer": exec_info(compatibility=31, core=31),
    "NamedFramebufferDrawBuffer": exec_info(compatibility=31, core=31),
    "NamedFramebufferDrawBuffers": exec_info(compatibility=31, core=31),
    "NamedFramebufferReadBuffer": exec_info(compatibility=31, core=31),
    "InvalidateNamedFramebufferData": exec_info(compatibility=31, core=31),
    "InvalidateNamedFramebufferSubData": exec_info(compatibility=31, core=31),
    "ClearNamedFramebufferiv": exec_info(compatibility=31, core=31),
    "ClearNamedFramebufferuiv": exec_info(compatibility=31, core=31),
    "ClearNamedFramebufferfv": exec_info(compatibility=31, core=31),
    "ClearNamedFramebufferfi": exec_info(compatibility=31, core=31),
    "BlitNamedFramebuffer": exec_info(compatibility=31, core=31),
    "CheckNamedFramebufferStatus": exec_info(compatibility=31, core=31),
    "GetNamedFramebufferParameteriv": exec_info(compatibility=31, core=31),
    "GetNamedFramebufferAttachmentParameteriv": exec_info(compatibility=31, core=31),
    "CreateRenderbuffers": exec_info(compatibility=31, core=31),
    "NamedRenderbufferStorage": exec_info(compatibility=31, core=31),
    "NamedRenderbufferStorageMultisample": exec_info(compatibility=31, core=31),
    "GetNamedRenderbufferParameteriv": exec_info(compatibility=31, core=31),
    "CreateTextures": exec_info(compatibility=31, core=31),
    "TextureBuffer": exec_info(compatibility=31, core=31),
    "TextureBufferRange": exec_info(compatibility=31, core=31),
    "TextureStorage1D": exec_info(compatibility=31, core=31),
    "TextureStorage2D": exec_info(compatibility=31, core=31),
    "TextureStorage3D": exec_info(compatibility=31, core=31),
    "TextureStorage2DMultisample": exec_info(compatibility=31, core=31),
    "TextureStorage3DMultisample": exec_info(compatibility=31, core=31),
    "TextureSubImage1D": exec_info(compatibility=31, core=31),
    "TextureSubImage2D": exec_info(compatibility=31, core=31),
    "TextureSubImage3D": exec_info(compatibility=31, core=31),
    "CompressedTextureSubImage1D": exec_info(compatibility=31, core=31),
    "CompressedTextureSubImage2D": exec_info(compatibility=31, core=31),
    "CompressedTextureSubImage3D": exec_info(compatibility=31, core=31),
    "CopyTextureSubImage1D": exec_info(compatibility=31, core=31),
    "CopyTextureSubImage2D": exec_info(compatibility=31, core=31),
    "CopyTextureSubImage3D": exec_info(compatibility=31, core=31),
    "TextureParameterf": exec_info(compatibility=31, core=31),
    "TextureParameterfv": exec_info(compatibility=31, core=31),
    "TextureParameteri": exec_info(compatibility=31, core=31),
    "TextureParameterIiv": exec_info(compatibility=31, core=31),
    "TextureParameterIuiv": exec_info(compatibility=31, core=31),
    "TextureParameteriv": exec_info(compatibility=31, core=31),
    "GenerateTextureMipmap": exec_info(compatibility=31, core=31),
    "BindTextureUnit": exec_info(compatibility=31, core=31),
    "GetTextureImage": exec_info(compatibility=31, core=31),
    "GetCompressedTextureImage": exec_info(compatibility=31, core=31),
    "GetTextureLevelParameterfv": exec_info(compatibility=31, core=31),
    "GetTextureLevelParameteriv": exec_info(compatibility=31, core=31),
    "GetTextureParameterfv": exec_info(compatibility=31, core=31),
    "GetTextureParameterIiv": exec_info(compatibility=31, core=31),
    "GetTextureParameterIuiv": exec_info(compatibility=31, core=31),
    "GetTextureParameteriv": exec_info(compatibility=31, core=31),
    "CreateVertexArrays": exec_info(compatibility=31, core=31),
    "DisableVertexArrayAttrib": exec_info(compatibility=31, core=31),
    "EnableVertexArrayAttrib": exec_info(compatibility=31, core=31),
    "VertexArrayElementBuffer": exec_info(compatibility=31, core=31),
    "VertexArrayVertexBuffer": exec_info(compatibility=31, core=31),
    "VertexArrayVertexBuffers": exec_info(compatibility=31, core=31),
    "VertexArrayAttribFormat": exec_info(compatibility=31, core=31),
    "VertexArrayAttribIFormat": exec_info(compatibility=31, core=31),
    "VertexArrayAttribLFormat": exec_info(compatibility=31, core=31),
    "VertexArrayAttribBinding": exec_info(compatibility=31, core=31),
    "VertexArrayBindingDivisor": exec_info(compatibility=31, core=31),
    "GetVertexArrayiv": exec_info(compatibility=31, core=31),
    "GetVertexArrayIndexediv": exec_info(compatibility=31, core=31),
    "GetVertexArrayIndexed64iv": exec_info(compatibility=31, core=31),
    "CreateSamplers": exec_info(compatibility=31, core=31),
    "CreateProgramPipelines": exec_info(compatibility=31, core=31),
    "CreateQueries": exec_info(compatibility=31, core=31),
    "GetQueryBufferObjectiv": exec_info(compatibility=31, core=31),
    "GetQueryBufferObjectuiv": exec_info(compatibility=31, core=31),
    "GetQueryBufferObjecti64v": exec_info(compatibility=31, core=31),
    "GetQueryBufferObjectui64v": exec_info(compatibility=31, core=31),

    # GL_ARB_gpu_shader_int64 - nominally requires OpenGL 4.0
    "Uniform1i64ARB": exec_info(compatibility=40, core=31),
    "Uniform2i64ARB": exec_info(compatibility=40, core=31),
    "Uniform3i64ARB": exec_info(compatibility=40, core=31),
    "Uniform4i64ARB": exec_info(compatibility=40, core=31),
    "Uniform1i64vARB": exec_info(compatibility=40, core=31),
    "Uniform2i64vARB": exec_info(compatibility=40, core=31),
    "Uniform3i64vARB": exec_info(compatibility=40, core=31),
    "Uniform4i64vARB": exec_info(compatibility=40, core=31),
    "Uniform1ui64ARB": exec_info(compatibility=40, core=31),
    "Uniform2ui64ARB": exec_info(compatibility=40, core=31),
    "Uniform3ui64ARB": exec_info(compatibility=40, core=31),
    "Uniform4ui64ARB": exec_info(compatibility=40, core=31),
    "Uniform1ui64vARB": exec_info(compatibility=40, core=31),
    "Uniform2ui64vARB": exec_info(compatibility=40, core=31),
    "Uniform3ui64vARB": exec_info(compatibility=40, core=31),
    "Uniform4ui64vARB": exec_info(compatibility=40, core=31),
    "GetUniformi64vARB": exec_info(compatibility=40, core=31),
    "GetUniformui64vARB": exec_info(compatibility=40, core=31),
    "GetnUniformi64vARB": exec_info(compatibility=40, core=31),
    "GetnUniformui64vARB": exec_info(compatibility=40, core=31),
    "ProgramUniform1i64ARB": exec_info(compatibility=40, core=31),
    "ProgramUniform2i64ARB": exec_info(compatibility=40, core=31),
    "ProgramUniform3i64ARB": exec_info(compatibility=40, core=31),
    "ProgramUniform4i64ARB": exec_info(compatibility=40, core=31),
    "ProgramUniform1i64vARB": exec_info(compatibility=40, core=31),
    "ProgramUniform2i64vARB": exec_info(compatibility=40, core=31),
    "ProgramUniform3i64vARB": exec_info(compatibility=40, core=31),
    "ProgramUniform4i64vARB": exec_info(compatibility=40, core=31),
    "ProgramUniform1ui64ARB": exec_info(compatibility=40, core=31),
    "ProgramUniform2ui64ARB": exec_info(compatibility=40, core=31),
    "ProgramUniform3ui64ARB": exec_info(compatibility=40, core=31),
    "ProgramUniform4ui64ARB": exec_info(compatibility=40, core=31),
    "ProgramUniform1ui64vARB": exec_info(compatibility=40, core=31),
    "ProgramUniform2ui64vARB": exec_info(compatibility=40, core=31),
    "ProgramUniform3ui64vARB": exec_info(compatibility=40, core=31),
    "ProgramUniform4ui64vARB": exec_info(compatibility=40, core=31),

    # GL_ARB_bindless_texture
    "GetVertexAttribLui64vARB": exec_info(compatibility=30, core=31),
}

def get_api_condition(f):
    condition_parts = []
    if f.name in functions:
        ex = functions[f.name]
        unconditional_count = 0

        if ex.compatibility is not None:
            condition_parts.append('ctx->API == API_OPENGL_COMPAT')
            unconditional_count += 1

        if ex.core is not None:
            condition_parts.append('ctx->API == API_OPENGL_CORE')
            unconditional_count += 1

        if ex.es1 is not None:
            condition_parts.append('ctx->API == API_OPENGLES')
            unconditional_count += 1

        if ex.es2 is not None:
            if ex.es2 > 20:
                condition_parts.append('(ctx->API == API_OPENGLES2 && ctx->Version >= {0})'.format(ex.es2))
            else:
                condition_parts.append('ctx->API == API_OPENGLES2')
                unconditional_count += 1

        # If the function is unconditionally available in all four
        # APIs, then it is always available.  Replace the complex
        # tautology condition with "true" and let GCC do the right
        # thing.
        if unconditional_count == 4:
            condition_parts = ['true']
    else:
        if f.desktop:
            if f.deprecated:
                condition_parts.append('ctx->API == API_OPENGL_COMPAT')
            else:
                condition_parts.append('_mesa_is_desktop_gl(ctx)')
        if 'es1' in f.api_map:
            condition_parts.append('ctx->API == API_OPENGLES')
        if 'es2' in f.api_map:
            if f.api_map['es2'] > 2.0:
                condition_parts.append('(ctx->API == API_OPENGLES2 && ctx->Version >= {0})'.format(int(f.api_map['es2'] * 10)))
            else:
                condition_parts.append('ctx->API == API_OPENGLES2')

    if not condition_parts:
        # This function does not exist in any API.
        return None
    return ' || '.join(condition_parts)

def print_glapi_file(printer):
    parser = argparse.ArgumentParser()
    parser.add_argument('-f',
                        dest='filename',
                        default='gl_and_es_API.xml',
                        help='an xml file describing an API')
    args = parser.parse_args()
    api = gl_XML.parse_GL_API(args.filename)
    printer.Print(api)




# Copyright (C) 2012 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

# This script generates the file api_exec_init.c, which contains
# _mesa_init_dispatch().  It is responsible for populating all
# entries in the "OutsideBeginEnd" dispatch table.

import argparse
import collections
import license
import gl_XML
import sys
import apiexec


exec_flavor_map = {
    'beginend': None,
    'dlist': '_mesa_',
    'mesa': '_mesa_',
    'skip': None,
    }


header = """/**
 * \\file api_exec_init.c
 * Initialize dispatch table.
 */


#include "api_exec_decl.h"
#include "glapi/glapi.h"
#include "main/context.h"
#include "main/dispatch.h"


/**
 * Initialize a context's OutsideBeginEnd table with pointers to Mesa's supported
 * GL functions.
 *
 * This function depends on ctx->Version.
 *
 * \param ctx  GL context
 */
void
_mesa_init_dispatch(struct gl_context *ctx)
{
   struct _glapi_table *table = ctx->OutsideBeginEnd;

   assert(table != NULL);
   assert(ctx->Version > 0);
"""


footer = """
}
"""


class PrintCode(gl_XML.gl_print_base):

    def __init__(self):
        gl_XML.gl_print_base.__init__(self)

        self.name = 'api_exec_init.py'
        self.license = license.bsd_license_template % (
            'Copyright (C) 2012 Intel Corporation',
            'Intel Corporation')

    def printRealHeader(self):
        print(header)

    def printRealFooter(self):
        print(footer)

    def printBody(self, api):
        # Collect SET_* calls by the condition under which they should
        # be called.
        settings_by_condition = collections.defaultdict(lambda: [])
        for f in api.functionIterateAll():
            if f.exec_flavor not in exec_flavor_map:
                raise Exception(
                    'Unrecognized exec flavor {0!r}'.format(f.exec_flavor))
            condition = apiexec.get_api_condition(f)
            if not condition:
                continue
            prefix = exec_flavor_map[f.exec_flavor]
            if prefix is None:
                # This function is not implemented, or is dispatched
                # via beginend.
                continue
            if f.has_no_error_variant:
                no_error_condition = '_mesa_is_no_error_enabled(ctx) && ({0})'.format(condition)
                error_condition = '!_mesa_is_no_error_enabled(ctx) && ({0})'.format(condition)
                settings_by_condition[no_error_condition].append(
                    'SET_{0}(table, {1}{0}_no_error);'.format(f.name, prefix, f.name))
                settings_by_condition[error_condition].append(
                    'SET_{0}(table, {1}{0});'.format(f.name, prefix, f.name))
            else:
                settings_by_condition[condition].append(
                    'SET_{0}(table, {1}{0});'.format(f.name, prefix, f.name))
        # Print out an if statement for each unique condition, with
        # the SET_* calls nested inside it.
        for condition in sorted(settings_by_condition.keys()):
            print('   if ({0}) {{'.format(condition))
            for setting in sorted(settings_by_condition[condition]):
                print('      {0}'.format(setting))
            print('   }')


if __name__ == '__main__':
    apiexec.print_glapi_file(PrintCode())



#encoding=utf-8
#
# Copyright © 2020 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

# A few utility functions reused across genxml scripts

import re

alphanum_nono = re.compile(r'[ /\[\]()\-:.,=>#&*\'"+\\]+')
def to_alphanum(name):
    global alphanum_nono
    return alphanum_nono.sub('', name)

def safe_name(name):
    name = to_alphanum(name)
    if not name[0].isalpha():
        name = '_' + name
    return name



COPYRIGHT = """\
/*
 * Copyright (C) 2017 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
"""

import argparse
import json
from sys import stdout
from mako.template import Template

def collect_data(spirv, kind):
    for x in spirv["operand_kinds"]:
        if x["kind"] == kind:
            operands = x
            break

    # There are some duplicate values in some of the tables (thanks guys!), so
    # filter them out.
    seen = set()
    values = []
    for x in operands["enumerants"]:
        if x["value"] not in seen:
            seen.add(x["value"])
            values.append(x["enumerant"])

    return (kind, values, operands["category"])

def collect_opcodes(spirv):
    seen = set()
    values = []
    for x in spirv["instructions"]:
        # Handle aliases by choosing the first one in the grammar.
        # E.g. OpDecorateString and OpDecorateStringGOOGLE share same opcode.
        if x["opcode"] in seen:
            continue
        opcode = x["opcode"]
        name = x["opname"]
        assert name.startswith("Op")
        values.append(name[2:])
        seen.add(opcode)

    return ("Op", values, None)

def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument("json")
    p.add_argument("out")
    return p.parse_args()

TEMPLATE  = Template("""\
/* DO NOT EDIT - This file is generated automatically by spirv_info_c.py script */

""" + COPYRIGHT + """\
#include "spirv_info.h"
% for kind,values,category in info:

% if category == "BitEnum":
const char *
spirv_${kind.lower()}_to_string(Spv${kind}Mask v)
{
   switch (v) {
    % for name in values:
    %if name != "None":
   case Spv${kind}${name}Mask: return "Spv${kind}${name}";
    % else:
   case Spv${kind}MaskNone: return "Spv${kind}${name}";
    % endif
    % endfor
   }

   return "unknown";
}
% else:
const char *
spirv_${kind.lower()}_to_string(Spv${kind} v)
{
   switch (v) {
    % for name in values:
   case Spv${kind}${name}: return "Spv${kind}${name}";
    % endfor
   case Spv${kind}Max: break; /* silence warnings about unhandled enums. */
   }

   return "unknown";
}
% endif
% endfor
""")

if __name__ == "__main__":
    pargs = parse_args()

    spirv_info = json.JSONDecoder().decode(open(pargs.json, "r").read())

    info = [
        collect_data(spirv_info, "AddressingModel"),
        collect_data(spirv_info, "BuiltIn"),
        collect_data(spirv_info, "Capability"),
        collect_data(spirv_info, "Decoration"),
        collect_data(spirv_info, "Dim"),
        collect_data(spirv_info, "ExecutionMode"),
        collect_data(spirv_info, "ExecutionModel"),
        collect_data(spirv_info, "ImageFormat"),
        collect_data(spirv_info, "MemoryModel"),
        collect_data(spirv_info, "StorageClass"),
        collect_data(spirv_info, "ImageOperands"),
        collect_data(spirv_info, "FPRoundingMode"),
        collect_opcodes(spirv_info),
    ]

    with open(pargs.out, 'w') as f:
        f.write(TEMPLATE.render(info=info))



#!/usr/bin/python3

import sys
import socket

host = sys.argv[1]
port = sys.argv[2]
mode = sys.argv[3]
relay = sys.argv[4]
msg = None

if mode == "on":
    msg = b'\x20'
else:
    msg = b'\x21'

msg += int(relay).to_bytes(1, 'big')
msg += b'\x00'

c = socket.create_connection((host, int(port)))
c.sendall(msg)

data = c.recv(1)
c.close()

if data[0] == b'\x01':
    print('Command failed')
    sys.exit(1)



#!/bin/env python
COPYRIGHT = """\
/*
 * Copyright 2021 Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sub license, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
 * IN NO EVENT SHALL VMWARE AND/OR ITS SUPPLIERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
"""

import argparse
import os.path
import re
import sys

from grl_parser import parse_grl_file

class Writer(object):
    def __init__(self, file):
        self._file = file
        self._indent = 0
        self._new_line = True

    def push_indent(self, levels=4):
        self._indent += levels

    def pop_indent(self, levels=4):
        self._indent -= levels

    def write(self, s, *fmt):
        if self._new_line:
            s = '\n' + s
        self._new_line = False
        if s.endswith('\n'):
            self._new_line = True
            s = s[:-1]
        if fmt:
            s = s.format(*fmt)
        self._file.write(s.replace('\n', '\n' + ' ' * self._indent))

# Internal Representation

class Value(object):
    def __init__(self, name=None, zone=None):
        self.name = name
        self._zone = zone
        self.live = False

    @property
    def zone(self):
        assert self._zone is not None
        return self._zone

    def is_reg(self):
        return False

    def c_val(self):
        if not self.name:
            print(self)
        assert self.name
        return self.name

    def c_cpu_val(self):
        assert self.zone == 'cpu'
        return self.c_val()

    def c_gpu_val(self):
        if self.zone == 'gpu':
            return self.c_val()
        else:
            return 'mi_imm({})'.format(self.c_cpu_val())

class Constant(Value):
    def __init__(self, value):
        super().__init__(zone='cpu')
        self.value = value

    def c_val(self):
        if self.value < 100:
            return str(self.value)
        elif self.value < (1 << 32):
            return '0x{:x}u'.format(self.value)
        else:
            return '0x{:x}ull'.format(self.value)

class Register(Value):
    def __init__(self, name):
        super().__init__(name=name, zone='gpu')

    def is_reg(self):
        return True

class FixedGPR(Register):
    def __init__(self, num):
        super().__init__('REG{}'.format(num))
        self.num = num

    def write_c(self, w):
        w.write('UNUSED struct mi_value {} = mi_reserve_gpr(&b, {});\n',
                self.name, self.num)

class GroupSizeRegister(Register):
    def __init__(self, comp):
        super().__init__('DISPATCHDIM_' + 'XYZ'[comp])
        self.comp = comp

class Member(Value):
    def __init__(self, value, member):
        super().__init__(zone=value.zone)
        self.value = value
        self.member = member

    def is_reg(self):
        return self.value.is_reg()

    def c_val(self):
        c_val = self.value.c_val()
        if self.zone == 'gpu':
            assert isinstance(self.value, Register)
            if self.member == 'hi':
                return 'mi_value_half({}, true)'.format(c_val)
            elif self.member == 'lo':
                return 'mi_value_half({}, false)'.format(c_val)
            else:
                assert False, 'Invalid member: {}'.format(self.member)
        else:
            return '.'.join([c_val, self.member])

class OffsetOf(Value):
    def __init__(self, mk, expr):
        super().__init__(zone='cpu')
        assert isinstance(expr, tuple) and expr[0] == 'member'
        self.type = mk.m.get_type(expr[1])
        self.field = expr[2]

    def c_val(self):
        return 'offsetof({}, {})'.format(self.type.c_name, self.field)

class Scope(object):
    def __init__(self, m, mk, parent):
        self.m = m
        self.mk = mk
        self.parent = parent
        self.defs = {}

    def add_def(self, d, name=None):
        if name is None:
            name = d.name
        assert name not in self.defs
        self.defs[name] = d

    def get_def(self, name):
        if name in self.defs:
            return self.defs[name]
        assert self.parent, 'Unknown definition: "{}"'.format(name)
        return self.parent.get_def(name)

class Statement(object):
    def __init__(self, srcs=[]):
        assert isinstance(srcs, (list, tuple))
        self.srcs = list(srcs)

class SSAStatement(Statement, Value):
    _count = 0

    def __init__(self, zone, srcs):
        Statement.__init__(self, srcs)
        Value.__init__(self, None, zone)
        self.c_name = '_tmp{}'.format(SSAStatement._count)
        SSAStatement._count += 1

    def c_val(self):
        return self.c_name

    def write_c_refs(self, w):
        assert self.zone == 'gpu'
        assert self.uses > 0
        if self.uses > 1:
            w.write('mi_value_add_refs(&b, {}, {});\n',
                    self.c_name, self.uses - 1)

class Half(SSAStatement):
    def __init__(self, value, half):
        assert half in ('hi', 'lo')
        super().__init__(None, [value])
        self.half = half

    @property
    def zone(self):
        return self.srcs[0].zone

    def write_c(self, w):
        assert self.half in ('hi', 'lo')
        if self.zone == 'cpu':
            if self.half == 'hi':
                w.write('uint32_t {} = (uint64_t)({}) >> 32;\n',
                        self.c_name, self.srcs[0].c_cpu_val())
            else:
                w.write('uint32_t {} = {};\n',
                        self.c_name, self.srcs[0].c_cpu_val())
        else:
            if self.half == 'hi':
                w.write('struct mi_value {} = mi_value_half({}, true);\n',
                        self.c_name, self.srcs[0].c_gpu_val())
            else:
                w.write('struct mi_value {} = mi_value_half({}, false);\n',
                        self.c_name, self.srcs[0].c_gpu_val())
            self.write_c_refs(w)

class Expression(SSAStatement):
    def __init__(self, mk, op, *srcs):
        super().__init__(None, srcs)
        self.op = op

    @property
    def zone(self):
        zone = 'cpu'
        for s in self.srcs:
            if s.zone == 'gpu':
                zone = 'gpu'
        return zone

    def write_c(self, w):
        if self.zone == 'cpu':
            w.write('uint64_t {} = ', self.c_name)
            c_cpu_vals = [s.c_cpu_val() for s in self.srcs]
            if len(self.srcs) == 1:
                w.write('({} {})', self.op, c_cpu_vals[0])
            elif len(self.srcs) == 2:
                w.write('({} {} {})', c_cpu_vals[0], self.op, c_cpu_vals[1])
            else:
                assert len(self.srcs) == 3 and op == '?'
                w.write('({} ? {} : {})', *c_cpu_vals)
            w.write(';\n')
            return

        w.write('struct mi_value {} = ', self.c_name)
        if self.op == '~':
            w.write('mi_inot(&b, {});\n', self.srcs[0].c_gpu_val())
        elif self.op == '+':
            w.write('mi_iadd(&b, {}, {});\n',
                    self.srcs[0].c_gpu_val(), self.srcs[1].c_gpu_val())
        elif self.op == '-':
            w.write('mi_isub(&b, {}, {});\n',
                    self.srcs[0].c_gpu_val(), self.srcs[1].c_gpu_val())
        elif self.op == '&':
            w.write('mi_iand(&b, {}, {});\n',
                    self.srcs[0].c_gpu_val(), self.srcs[1].c_gpu_val())
        elif self.op == '|':
            w.write('mi_ior(&b, {}, {});\n',
                    self.srcs[0].c_gpu_val(), self.srcs[1].c_gpu_val())
        elif self.op == '<<':
            if self.srcs[1].zone == 'cpu':
                w.write('mi_ishl_imm(&b, {}, {});\n',
                        self.srcs[0].c_gpu_val(), self.srcs[1].c_cpu_val())
            else:
                w.write('mi_ishl(&b, {}, {});\n',
                        self.srcs[0].c_gpu_val(), self.srcs[1].c_gpu_val())
        elif self.op == '>>':
            if self.srcs[1].zone == 'cpu':
                w.write('mi_ushr_imm(&b, {}, {});\n',
                        self.srcs[0].c_gpu_val(), self.srcs[1].c_cpu_val())
            else:
                w.write('mi_ushr(&b, {}, {});\n',
                        self.srcs[0].c_gpu_val(), self.srcs[1].c_gpu_val())
        elif self.op == '==':
            w.write('mi_ieq(&b, {}, {});\n',
                    self.srcs[0].c_gpu_val(), self.srcs[1].c_gpu_val())
        elif self.op == '<':
            w.write('mi_ult(&b, {}, {});\n',
                    self.srcs[0].c_gpu_val(), self.srcs[1].c_gpu_val())
        elif self.op == '>':
            w.write('mi_ult(&b, {}, {});\n',
                    self.srcs[1].c_gpu_val(), self.srcs[0].c_gpu_val())
        elif self.op == '<=':
            w.write('mi_uge(&b, {}, {});\n',
                    self.srcs[1].c_gpu_val(), self.srcs[0].c_gpu_val())
        else:
            assert False, 'Unknown expression opcode: {}'.format(self.op)
        self.write_c_refs(w)

class StoreReg(Statement):
    def __init__(self, mk, reg, value):
        super().__init__([mk.load_value(value)])
        self.reg = mk.parse_value(reg)
        assert self.reg.is_reg()

    def write_c(self, w):
        value = self.srcs[0]
        w.write('mi_store(&b, {}, {});\n',
                self.reg.c_gpu_val(), value.c_gpu_val())

class LoadMem(SSAStatement):
    def __init__(self, mk, bit_size, addr):
        super().__init__('gpu', [mk.load_value(addr)])
        self.bit_size = bit_size

    def write_c(self, w):
        addr = self.srcs[0]
        w.write('struct mi_value {} = ', self.c_name)
        if addr.zone == 'cpu':
            w.write('mi_mem{}(anv_address_from_u64({}));\n',
                    self.bit_size, addr.c_cpu_val())
        else:
            assert self.bit_size == 64
            w.write('mi_load_mem64_offset(&b, anv_address_from_u64(0), {});\n',
                    addr.c_gpu_val())
        self.write_c_refs(w)

class StoreMem(Statement):
    def __init__(self, mk, bit_size, addr, src):
        super().__init__([mk.load_value(addr), mk.load_value(src)])
        self.bit_size = bit_size

    def write_c(self, w):
        addr, data = tuple(self.srcs)
        if addr.zone == 'cpu':
            w.write('mi_store(&b, mi_mem{}(anv_address_from_u64({})), {});\n',
                    self.bit_size, addr.c_cpu_val(), data.c_gpu_val())
        else:
            assert self.bit_size == 64
            w.write('mi_store_mem64_offset(&b, anv_address_from_u64(0), {}, {});\n',
                    addr.c_gpu_val(), data.c_gpu_val())

class GoTo(Statement):
    def __init__(self, mk, target_id, cond=None, invert=False):
        cond = [mk.load_value(cond)] if cond is not None else []
        super().__init__(cond)
        self.target_id = target_id
        self.invert = invert
        self.mk = mk

    def write_c(self, w):
        # Now that we've parsed the entire metakernel, we can look up the
        # actual target from the id
        target = self.mk.get_goto_target(self.target_id)

        if self.srcs:
            cond = self.srcs[0]
            if self.invert:
                w.write('mi_goto_if(&b, mi_inot(&b, {}), &{});\n', cond.c_gpu_val(), target.c_name)
            else:
                w.write('mi_goto_if(&b, {}, &{});\n', cond.c_gpu_val(), target.c_name)
        else:
            w.write('mi_goto(&b, &{});\n', target.c_name)

class GoToTarget(Statement):
    def __init__(self, mk, name):
        super().__init__()
        self.name = name
        self.c_name = '_goto_target_' + name
        self.goto_tokens = []

        mk = mk.add_goto_target(self)

    def write_decl(self, w):
        w.write('struct mi_goto_target {} = MI_GOTO_TARGET_INIT;\n',
                self.c_name)

    def write_c(self, w):
        w.write('mi_goto_target(&b, &{});\n', self.c_name)

class Dispatch(Statement):
    def __init__(self, mk, kernel, group_size, args, postsync):
        if group_size is None:
            srcs = [mk.scope.get_def('DISPATCHDIM_{}'.format(d)) for d in 'XYZ']
        else:
            srcs = [mk.load_value(s) for s in group_size]
        srcs += [mk.load_value(a) for a in args]
        super().__init__(srcs)
        self.kernel = mk.m.kernels[kernel]
        self.indirect = group_size is None
        self.postsync = postsync

    def write_c(self, w):
        w.write('{\n')
        w.push_indent()

        group_size = self.srcs[:3]
        args = self.srcs[3:]
        if not self.indirect:
            w.write('const uint32_t _group_size[3] = {{ {}, {}, {} }};\n',
                    *[s.c_cpu_val() for s in group_size])
            gs = '_group_size'
        else:
            gs = 'NULL'

        w.write('const struct anv_kernel_arg _args[] = {\n')
        w.push_indent()
        for arg in args:
            w.write('{{ .u64 = {} }},\n', arg.c_cpu_val())
        w.pop_indent()
        w.write('};\n')

        w.write('genX(grl_dispatch)(cmd_buffer, {},\n', self.kernel.c_name)
        w.write('                   {}, ARRAY_SIZE(_args), _args);\n', gs)
        w.pop_indent()
        w.write('}\n')

class SemWait(Statement):
    def __init__(self, scope, wait):
        super().__init__()
        self.wait = wait

class Control(Statement):
    def __init__(self, scope, wait):
        super().__init__()
        self.wait = wait

    def write_c(self, w):
        w.write('cmd_buffer->state.pending_pipe_bits |=\n')
        w.write('    ANV_PIPE_CS_STALL_BIT |\n')
        w.write('    ANV_PIPE_DATA_CACHE_FLUSH_BIT |\n')
        w.write('    ANV_PIPE_UNTYPED_DATAPORT_CACHE_FLUSH_BIT;\n')
        w.write('genX(cmd_buffer_apply_pipe_flushes)(cmd_buffer);\n')

TYPE_REMAPS = {
    'dword' : 'uint32_t',
    'qword' : 'uint64_t',
}

class Module(object):
    def __init__(self, grl_dir, elems):
        assert isinstance(elems[0], tuple)
        assert elems[0][0] == 'module-name'
        self.grl_dir = grl_dir
        self.name = elems[0][1]
        self.kernels = {}
        self.structs = {}
        self.constants = []
        self.metakernels = []
        self.regs = {}

        scope = Scope(self, None, None)
        for e in elems[1:]:
            if e[0] == 'kernel':
                k = Kernel(self, *e[1:])
                assert k.name not in self.kernels
                self.kernels[k.name] = k
            elif e[0] == 'kernel-module':
                m = KernelModule(self, *e[1:])
                for k in m.kernels:
                    assert k.name not in self.kernels
                    self.kernels[k.name] = k
            elif e[0] == 'struct':
                s = Struct(self, *e[1:])
                assert s.name not in self.kernels
                self.structs[s.name] = s
            elif e[0] == 'named-constant':
                c = NamedConstant(*e[1:])
                scope.add_def(c)
                self.constants.append(c)
            elif e[0] == 'meta-kernel':
                mk = MetaKernel(self, scope, *e[1:])
                self.metakernels.append(mk)
            elif e[0] == 'import':
                assert e[2] == 'struct'
                self.import_struct(e[1], e[3])
            else:
                assert False, 'Invalid module-level token: {}'.format(t[0])

    def import_struct(self, filename, struct_name):
        elems = parse_grl_file(os.path.join(self.grl_dir, filename), [])
        assert elems
        for e in elems[1:]:
            if e[0] == 'struct' and e[1] == struct_name:
                s = Struct(self, *e[1:])
                assert s.name not in self.kernels
                self.structs[s.name] = s
                return
        assert False, "Struct {0} not found in {1}".format(struct_name, filename)

    def get_type(self, name):
        if name in self.structs:
            return self.structs[name]
        return BasicType(TYPE_REMAPS.get(name, name))

    def get_fixed_gpr(self, num):
        assert isinstance(num, int)
        if num in self.regs:
            return self.regs[num]

        reg = FixedGPR(num)
        self.regs[num] = reg
        return reg

    def optimize(self):
        progress = True
        while progress:
            progress = False

            # Copy Propagation
            for mk in self.metakernels:
                if mk.opt_copy_prop():
                    progress = True

            # Dead Code Elimination
            for r in self.regs.values():
                r.live = False
            for c in self.constants:
                c.live = False
            for mk in self.metakernels:
                mk.opt_dead_code1()
            for mk in self.metakernels:
                if mk.opt_dead_code2():
                    progress = True
            for n in list(self.regs.keys()):
                if not self.regs[n].live:
                    del self.regs[n]
                    progress = True
            self.constants = [c for c in self.constants if c.live]

    def compact_regs(self):
        old_regs = self.regs
        self.regs = {}
        for i, reg in enumerate(old_regs.values()):
            reg.num = i
            self.regs[i] = reg

    def write_h(self, w):
        for s in self.structs.values():
            s.write_h(w)
        for mk in self.metakernels:
            mk.write_h(w)

    def write_c(self, w):
        for c in self.constants:
            c.write_c(w)
        for mk in self.metakernels:
            mk.write_c(w)

class Kernel(object):
    def __init__(self, m, name, ann):
        self.name = name
        self.source_file = ann['source']
        self.kernel_name = self.source_file.replace('/', '_')[:-3].upper()
        self.entrypoint = ann['kernelFunction']

        assert self.source_file.endswith('.cl')
        self.c_name = '_'.join([
            'GRL_CL_KERNEL',
            self.kernel_name,
            self.entrypoint.upper(),
        ])

class KernelModule(object):
    def __init__(self, m, name, source, kernels):
        self.name = name
        self.kernels = []
        self.libraries = []

        for k in kernels:
            if k[0] == 'kernel':
                k[2]['source'] = source
                self.kernels.append(Kernel(m, *k[1:]))
            elif k[0] == 'library':
                # Skip this for now.
                pass

class BasicType(object):
    def __init__(self, name):
        self.name = name
        self.c_name = name

class Struct(object):
    def __init__(self, m, name, fields, align):
        assert align == 0
        self.name = name
        self.c_name = 'struct ' + '_'.join(['grl', m.name, self.name])
        self.fields = [(m.get_type(t), n) for t, n in fields]

    def write_h(self, w):
        w.write('{} {{\n', self.c_name)
        w.push_indent()
        for f in self.fields:
            w.write('{} {};\n', f[0].c_name, f[1])
        w.pop_indent()
        w.write('};\n')

class NamedConstant(Value):
    def __init__(self, name, value):
        super().__init__(name, 'cpu')
        self.name = name
        self.value = Constant(value)
        self.written = False

    def set_module(self, m):
        pass

    def write_c(self, w):
        if self.written:
            return
        w.write('static const uint64_t {} = {};\n',
                self.name, self.value.c_val())
        self.written = True

class MetaKernelParameter(Value):
    def __init__(self, mk, type, name):
        super().__init__(name, 'cpu')
        self.type = mk.m.get_type(type)

class MetaKernel(object):
    def __init__(self, m, m_scope, name, params, ann, statements):
        self.m = m
        self.name = name
        self.c_name = '_'.join(['grl', m.name, self.name])
        self.goto_targets = {}
        self.num_tmps = 0

        mk_scope = Scope(m, self, m_scope)

        self.params = [MetaKernelParameter(self, *p) for p in params]
        for p in self.params:
            mk_scope.add_def(p)

        mk_scope.add_def(GroupSizeRegister(0), name='DISPATCHDIM_X')
        mk_scope.add_def(GroupSizeRegister(1), name='DISPATCHDIM_Y')
        mk_scope.add_def(GroupSizeRegister(2), name='DISPATCHDIM_Z')

        self.statements = []
        self.parse_stmt(mk_scope, statements)
        self.scope = None

    def get_tmp(self):
        tmpN = '_tmp{}'.format(self.num_tmps)
        self.num_tmps += 1
        return tmpN

    def add_stmt(self, stmt):
        self.statements.append(stmt)
        return stmt

    def parse_value(self, v):
        if isinstance(v, Value):
            return v
        elif isinstance(v, str):
            if re.match(r'REG\d+', v):
                return self.m.get_fixed_gpr(int(v[3:]))
            else:
                return self.scope.get_def(v)
        elif isinstance(v, int):
            return Constant(v)
        elif isinstance(v, tuple):
            if v[0] == 'member':
                return Member(self.parse_value(v[1]), v[2])
            elif v[0] == 'offsetof':
                return OffsetOf(self, v[1])
            else:
                op = v[0]
                srcs = [self.parse_value(s) for s in v[1:]]
                return self.add_stmt(Expression(self, op, *srcs))
        else:
            assert False, 'Invalid value: {}'.format(v[0])

    def load_value(self, v):
        v = self.parse_value(v)
        if isinstance(v, Member) and v.zone == 'gpu':
            v = self.add_stmt(Half(v.value, v.member))
        return v

    def parse_stmt(self, scope, s):
        self.scope = scope
        if isinstance(s, list):
            subscope = Scope(self.m, self, scope)
            for stmt in s:
                self.parse_stmt(subscope, stmt)
        elif s[0] == 'define':
            scope.add_def(self.parse_value(s[2]), name=s[1])
        elif s[0] == 'assign':
            self.add_stmt(StoreReg(self, *s[1:]))
        elif s[0] == 'dispatch':
            self.add_stmt(Dispatch(self, *s[1:]))
        elif s[0] == 'load-dword':
            v = self.add_stmt(LoadMem(self, 32, s[2]))
            self.add_stmt(StoreReg(self, s[1], v))
        elif s[0] == 'load-qword':
            v = self.add_stmt(LoadMem(self, 64, s[2]))
            self.add_stmt(StoreReg(self, s[1], v))
        elif s[0] == 'store-dword':
            self.add_stmt(StoreMem(self, 32, *s[1:]))
        elif s[0] == 'store-qword':
            self.add_stmt(StoreMem(self, 64, *s[1:]))
        elif s[0] == 'goto':
            self.add_stmt(GoTo(self, s[1]))
        elif s[0] == 'goto-if':
            self.add_stmt(GoTo(self, s[1], s[2]))
        elif s[0] == 'goto-if-not':
            self.add_stmt(GoTo(self, s[1], s[2], invert=True))
        elif s[0] == 'label':
            self.add_stmt(GoToTarget(self, s[1]))
        elif s[0] == 'control':
            self.add_stmt(Control(self, s[1]))
        elif s[0] == 'sem-wait-while':
            self.add_stmt(Control(self, s[1]))
        else:
            assert False, 'Invalid statement: {}'.format(s[0])

    def add_goto_target(self, t):
        assert t.name not in self.goto_targets
        self.goto_targets[t.name] = t

    def get_goto_target(self, name):
        return self.goto_targets[name]

    def opt_copy_prop(self):
        progress = False
        copies = {}
        for stmt in self.statements:
            for i in range(len(stmt.srcs)):
                src = stmt.srcs[i]
                if isinstance(src, FixedGPR) and src.num in copies:
                    stmt.srcs[i] = copies[src.num]
                    progress = True

            if isinstance(stmt, StoreReg):
                reg = stmt.reg
                if isinstance(reg, Member):
                    reg = reg.value

                if isinstance(reg, FixedGPR):
                    copies.pop(reg.num, None)
                    if not stmt.srcs[0].is_reg():
                        copies[reg.num] = stmt.srcs[0]
            elif isinstance(stmt, (GoTo, GoToTarget)):
                copies = {}

        return progress

    def opt_dead_code1(self):
        for stmt in self.statements:
            # Mark every register which is read as live
            for src in stmt.srcs:
                if isinstance(src, Register):
                    src.live = True

            # Initialize every SSA statement to dead
            if isinstance(stmt, SSAStatement):
                stmt.live = False

    def opt_dead_code2(self):
        def yield_live(statements):
            gprs_read = set(self.m.regs.keys())
            for stmt in statements:
                if isinstance(stmt, SSAStatement):
                    if not stmt.live:
                        continue
                elif isinstance(stmt, StoreReg):
                    reg = stmt.reg
                    if isinstance(reg, Member):
                        reg = reg.value

                    if not stmt.reg.live:
                        continue

                    if isinstance(reg, FixedGPR):
                        if reg.num in gprs_read:
                            gprs_read.remove(reg.num)
                        else:
                            continue
                elif isinstance(stmt, (GoTo, GoToTarget)):
                    gprs_read = set(self.m.regs.keys())

                for src in stmt.srcs:
                    src.live = True
                    if isinstance(src, FixedGPR):
                        gprs_read.add(src.num)
                yield stmt

        old_stmt_list = self.statements
        old_stmt_list.reverse()
        self.statements = list(yield_live(old_stmt_list))
        self.statements.reverse()
        return len(self.statements) != len(old_stmt_list)

    def count_ssa_value_uses(self):
        for stmt in self.statements:
            if isinstance(stmt, SSAStatement):
                stmt.uses = 0

            for src in stmt.srcs:
                if isinstance(src, SSAStatement):
                    src.uses += 1

    def write_h(self, w):
        w.write('void\n')
        w.write('genX({})(\n', self.c_name)
        w.push_indent()
        w.write('struct anv_cmd_buffer *cmd_buffer')
        for p in self.params:
            w.write(',\n{} {}', p.type.c_name, p.name)
        w.write(');\n')
        w.pop_indent()

    def write_c(self, w):
        w.write('void\n')
        w.write('genX({})(\n', self.c_name)
        w.push_indent()
        w.write('struct anv_cmd_buffer *cmd_buffer')
        for p in self.params:
            w.write(',\n{} {}', p.type.c_name, p.name)
        w.write(')\n')
        w.pop_indent()
        w.write('{\n')
        w.push_indent()

        w.write('struct mi_builder b;\n')
        w.write('mi_builder_init(&b, cmd_buffer->device->info, &cmd_buffer->batch);\n')
        w.write('\n')

        for r in self.m.regs.values():
            r.write_c(w)
        w.write('\n')

        for t in self.goto_targets.values():
            t.write_decl(w)
        w.write('\n')

        self.count_ssa_value_uses()
        for s in self.statements:
            s.write_c(w)

        w.pop_indent()

        w.write('}\n')

HEADER_PROLOGUE = COPYRIGHT + '''
#include "anv_private.h"
#include "grl/genX_grl.h"

#ifndef {0}
#define {0}

#ifdef __cplusplus
extern "C" {{
#endif

'''

HEADER_EPILOGUE = '''
#ifdef __cplusplus
}}
#endif

#endif /* {0} */
'''

C_PROLOGUE = COPYRIGHT + '''
#include "{0}"

#include "genxml/gen_macros.h"
#include "genxml/genX_pack.h"
#include "genxml/gen_rt_pack.h"

/* We reserve :
 *    - GPR 14 for secondary command buffer returns
 *    - GPR 15 for conditional rendering
 */
#define MI_BUILDER_NUM_ALLOC_GPRS 14
#define __gen_get_batch_dwords anv_batch_emit_dwords
#define __gen_address_offset anv_address_add
#define __gen_get_batch_address(b, a) anv_batch_address(b, a)
#include "common/mi_builder.h"

#define MI_PREDICATE_RESULT mi_reg32(0x2418)
#define DISPATCHDIM_X mi_reg32(0x2500)
#define DISPATCHDIM_Y mi_reg32(0x2504)
#define DISPATCHDIM_Z mi_reg32(0x2508)
'''

def parse_libraries(filenames):
    libraries = {}
    for fname in filenames:
        lib_package = parse_grl_file(fname, [])
        for lib in lib_package:
            assert lib[0] == 'library'
            # Add the directory of the library so that CL files can be found.
            lib[2].append(('path', os.path.dirname(fname)))
            libraries[lib[1]] = lib
    return libraries

def main():
    argparser = argparse.ArgumentParser()
    argparser.add_argument('--out-c', help='Output C file')
    argparser.add_argument('--out-h', help='Output C file')
    argparser.add_argument('--library', dest='libraries', action='append',
                           default=[], help='Libraries to include')
    argparser.add_argument('grl', help="Input  file")
    args = argparser.parse_args()

    grl_dir = os.path.dirname(args.grl)

    libraries = parse_libraries(args.libraries)

    ir = parse_grl_file(args.grl, libraries)

    m = Module(grl_dir, ir)
    m.optimize()
    m.compact_regs()

    with open(args.out_h, 'w') as f:
        guard = os.path.splitext(os.path.basename(args.out_h))[0].upper()
        w = Writer(f)
        w.write(HEADER_PROLOGUE, guard)
        m.write_h(w)
        w.write(HEADER_EPILOGUE, guard)

    with open(args.out_c, 'w') as f:
        w = Writer(f)
        w.write(C_PROLOGUE, os.path.basename(args.out_h))
        m.write_c(w)

if __name__ == '__main__':
    main()



# Copyright © 2021 Igalia S.L.
# SPDX-License-Identifier: MIT

import argparse
import sys

#
# TODO can we do this with less boilerplate?
#
parser = argparse.ArgumentParser()
parser.add_argument('-p', '--import-path', required=True)
parser.add_argument('--utrace-src', required=True)
parser.add_argument('--utrace-hdr', required=True)
parser.add_argument('--perfetto-hdr', required=True)
args = parser.parse_args()
sys.path.insert(0, args.import_path)


from u_trace import Header, HeaderScope
from u_trace import ForwardDecl
from u_trace import Tracepoint
from u_trace import TracepointArg as Arg
from u_trace import TracepointArgStruct as ArgStruct
from u_trace import utrace_generate
from u_trace import utrace_generate_perfetto_utils

Header('vk_enum_to_str.h', scope=HeaderScope.SOURCE|HeaderScope.PERFETTO)
Header('vk_format.h')
Header('tu_cmd_buffer.h', scope=HeaderScope.SOURCE)
Header('tu_device.h', scope=HeaderScope.SOURCE)

# we can't use tu_common.h because it includes ir3 headers which are not
# compatible with C++
ForwardDecl('struct tu_cmd_buffer')
ForwardDecl('struct tu_device')
ForwardDecl('struct tu_framebuffer')
ForwardDecl('struct tu_tiling_config')

# List of the default tracepoints enabled. By default tracepoints are enabled,
# set tp_default_enabled=False to disable them by default.
tu_default_tps = []

#
# Tracepoint definitions:
#

def begin_end_tp(name, args=[], tp_struct=None, tp_print=None,
                 tp_default_enabled=True):
    global tu_default_tps
    if tp_default_enabled:
        tu_default_tps.append(name)
    Tracepoint('start_{0}'.format(name),
               toggle_name=name,
               tp_perfetto='tu_start_{0}'.format(name))
    Tracepoint('end_{0}'.format(name),
               toggle_name=name,
               args=args,
               tp_struct=tp_struct,
               tp_perfetto='tu_end_{0}'.format(name),
               tp_print=tp_print)

begin_end_tp('cmd_buffer',
    args=[ArgStruct(type='const struct tu_cmd_buffer *', var='cmd')],
    tp_struct=[Arg(type='VkCommandBufferLevel', name='level', var='cmd->vk.level', c_format='%s', to_prim_type='vk_CommandBufferLevel_to_str({})'),
               Arg(type='uint8_t', name='render_pass_continue', var='!!(cmd->usage_flags & VK_COMMAND_BUFFER_USAGE_RENDER_PASS_CONTINUE_BIT)', c_format='%u')])

begin_end_tp('render_pass',
    args=[ArgStruct(type='const struct tu_framebuffer *', var='fb'),
          ArgStruct(type='const struct tu_tiling_config *', var='tiling')],
    tp_struct=[Arg(type='uint16_t', name='width',        var='fb->width',                                    c_format='%u'),
               Arg(type='uint16_t', name='height',       var='fb->height',                                   c_format='%u'),
               Arg(type='uint8_t',  name='MRTs',         var='fb->attachment_count',                         c_format='%u'),
            #    Arg(type='uint8_t',  name='samples',      var='fb->samples',                                  c_format='%u'),
               Arg(type='uint16_t', name='numberOfBins', var='tiling->tile_count.width * tiling->tile_count.height', c_format='%u'),
               Arg(type='uint16_t', name='binWidth',     var='tiling->tile0.width',                                  c_format='%u'),
               Arg(type='uint16_t', name='binHeight',    var='tiling->tile0.height',                                 c_format='%u')])

begin_end_tp('binning_ib')
begin_end_tp('draw_ib_sysmem')
begin_end_tp('draw_ib_gmem')

begin_end_tp('gmem_clear',
    args=[Arg(type='enum VkFormat',  var='format',  c_format='%s', to_prim_type='vk_format_description({})->short_name'),
          Arg(type='uint8_t',        var='samples', c_format='%u')])

begin_end_tp('sysmem_clear',
    args=[Arg(type='enum VkFormat',  var='format',      c_format='%s', to_prim_type='vk_format_description({})->short_name'),
          Arg(type='uint8_t',        var='uses_3d_ops', c_format='%u'),
          Arg(type='uint8_t',        var='samples',     c_format='%u')])

begin_end_tp('sysmem_clear_all',
    args=[Arg(type='uint8_t',        var='mrt_count',   c_format='%u'),
          Arg(type='uint8_t',        var='rect_count',  c_format='%u')])

begin_end_tp('gmem_load',
    args=[Arg(type='enum VkFormat',  var='format',   c_format='%s', to_prim_type='vk_format_description({})->short_name'),
          Arg(type='uint8_t',        var='force_load', c_format='%u')])

begin_end_tp('gmem_store',
    args=[Arg(type='enum VkFormat',  var='format',   c_format='%s', to_prim_type='vk_format_description({})->short_name'),
          Arg(type='uint8_t',        var='fast_path', c_format='%u'),
          Arg(type='uint8_t',        var='unaligned', c_format='%u')])

begin_end_tp('sysmem_resolve',
    args=[Arg(type='enum VkFormat',  var='format',   c_format='%s', to_prim_type='vk_format_description({})->short_name')])

begin_end_tp('blit',
    # TODO: add source megapixels count and target megapixels count arguments
    args=[Arg(type='uint8_t',        var='uses_3d_blit', c_format='%u'),
          Arg(type='enum VkFormat',  var='src_format',   c_format='%s', to_prim_type='vk_format_description({})->short_name'),
          Arg(type='enum VkFormat',  var='dst_format',   c_format='%s', to_prim_type='vk_format_description({})->short_name'),
          Arg(type='uint8_t',        var='layers',       c_format='%u')])

begin_end_tp('compute',
    args=[Arg(type='uint8_t',  var='indirect',       c_format='%u'),
          Arg(type='uint16_t', var='local_size_x',   c_format='%u'),
          Arg(type='uint16_t', var='local_size_y',   c_format='%u'),
          Arg(type='uint16_t', var='local_size_z',   c_format='%u'),
          Arg(type='uint16_t', var='num_groups_x',   c_format='%u'),
          Arg(type='uint16_t', var='num_groups_y',   c_format='%u'),
          Arg(type='uint16_t', var='num_groups_z',   c_format='%u')])

utrace_generate(cpath=args.utrace_src,
                hpath=args.utrace_hdr,
                ctx_param='struct tu_device *dev',
                trace_toggle_name='tu_gpu_tracepoint',
                trace_toggle_defaults=tu_default_tps)
utrace_generate_perfetto_utils(hpath=args.perfetto_hdr)



# Copyright 2020 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sub license, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice (including the
# next paragraph) shall be included in all copies or substantial portions
# of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
# IN NO EVENT SHALL VMWARE AND/OR ITS SUPPLIERS BE LIABLE FOR
# ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import xml.etree.ElementTree as et

from collections import OrderedDict, namedtuple

# Mesa-local imports must be declared in meson variable
# '{file_without_suffix}_depend_files'.
from vk_extensions import Extension, VkVersion

EntrypointParam = namedtuple('EntrypointParam', 'type name decl len')

class EntrypointBase:
    def __init__(self, name):
        assert name.startswith('vk')
        self.name = name[2:]
        self.alias = None
        self.guard = None
        self.entry_table_index = None
        # Extensions which require this entrypoint
        self.core_version = None
        self.extensions = []

    def prefixed_name(self, prefix):
        return prefix + '_' + self.name

class Entrypoint(EntrypointBase):
    def __init__(self, name, return_type, params, guard=None):
        super(Entrypoint, self).__init__(name)
        self.return_type = return_type
        self.params = params
        self.guard = guard
        self.aliases = []
        self.disp_table_index = None

    def is_physical_device_entrypoint(self):
        return self.params[0].type in ('VkPhysicalDevice', )

    def is_device_entrypoint(self):
        return self.params[0].type in ('VkDevice', 'VkCommandBuffer', 'VkQueue')

    def decl_params(self, start=0):
        return ', '.join(p.decl for p in self.params[start:])

    def call_params(self, start=0):
        return ', '.join(p.name for p in self.params[start:])

class EntrypointAlias(EntrypointBase):
    def __init__(self, name, entrypoint):
        super(EntrypointAlias, self).__init__(name)
        self.alias = entrypoint
        entrypoint.aliases.append(self)

    def is_physical_device_entrypoint(self):
        return self.alias.is_physical_device_entrypoint()

    def is_device_entrypoint(self):
        return self.alias.is_device_entrypoint()

    def prefixed_name(self, prefix):
        return self.alias.prefixed_name(prefix)

    @property
    def params(self):
        return self.alias.params

    @property
    def return_type(self):
        return self.alias.return_type

    @property
    def disp_table_index(self):
        return self.alias.disp_table_index

    def decl_params(self):
        return self.alias.decl_params()

    def call_params(self):
        return self.alias.call_params()

def get_entrypoints(doc, entrypoints_to_defines):
    """Extract the entry points from the registry."""
    entrypoints = OrderedDict()

    for command in doc.findall('./commands/command'):
        if 'alias' in command.attrib:
            alias = command.attrib['name']
            target = command.attrib['alias']
            entrypoints[alias] = EntrypointAlias(alias, entrypoints[target])
        else:
            name = command.find('./proto/name').text
            ret_type = command.find('./proto/type').text
            params = [EntrypointParam(
                type=p.find('./type').text,
                name=p.find('./name').text,
                decl=''.join(p.itertext()),
                len=p.attrib.get('altlen', p.attrib.get('len', None))
            ) for p in command.findall('./param')]
            guard = entrypoints_to_defines.get(name)
            # They really need to be unique
            assert name not in entrypoints
            entrypoints[name] = Entrypoint(name, ret_type, params, guard)

    for feature in doc.findall('./feature'):
        assert feature.attrib['api'] == 'vulkan'
        version = VkVersion(feature.attrib['number'])
        for command in feature.findall('./require/command'):
            e = entrypoints[command.attrib['name']]
            assert e.core_version is None
            e.core_version = version

    for extension in doc.findall('.extensions/extension'):
        if extension.attrib['supported'] != 'vulkan':
            continue

        ext_name = extension.attrib['name']

        ext = Extension(ext_name, 1, True)
        ext.type = extension.attrib['type']

        for command in extension.findall('./require/command'):
            e = entrypoints[command.attrib['name']]
            assert e.core_version is None
            e.extensions.append(ext)

    return entrypoints.values()


def get_entrypoints_defines(doc):
    """Maps entry points to extension defines."""
    entrypoints_to_defines = {}

    platform_define = {}
    for platform in doc.findall('./platforms/platform'):
        name = platform.attrib['name']
        define = platform.attrib['protect']
        platform_define[name] = define

    for extension in doc.findall('./extensions/extension[@platform]'):
        platform = extension.attrib['platform']
        define = platform_define[platform]

        for entrypoint in extension.findall('./require/command'):
            fullname = entrypoint.attrib['name']
            entrypoints_to_defines[fullname] = define

    return entrypoints_to_defines

def get_entrypoints_from_xml(xml_files):
    entrypoints = []

    for filename in xml_files:
        doc = et.parse(filename)
        entrypoints += get_entrypoints(doc, get_entrypoints_defines(doc))

    return entrypoints




template = """\
/* Copyright (C) 2018 Red Hat
 * Copyright (C) 2020 Valve Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef _NIR_INTRINSICS_INDICES_
#define _NIR_INTRINSICS_INDICES_

% for index in INTR_INDICES:
<%
data_type = index.c_data_type
name = index.name
enum = "NIR_INTRINSIC_" + name.upper()
%>

static inline ${data_type}
nir_intrinsic_${name}(const nir_intrinsic_instr *instr)
{
   const nir_intrinsic_info *info = &nir_intrinsic_infos[instr->intrinsic];
   assert(info->index_map[${enum}] > 0);
% if "struct" in data_type:
   ${data_type} res;
   STATIC_ASSERT(sizeof(instr->const_index[0]) == sizeof(res));
   memcpy(&res, &instr->const_index[info->index_map[${enum}] - 1], sizeof(res));
   return res;
% else:
   return (${data_type})instr->const_index[info->index_map[${enum}] - 1];
% endif
}

static inline void
nir_intrinsic_set_${name}(nir_intrinsic_instr *instr, ${data_type} val)
{
   const nir_intrinsic_info *info = &nir_intrinsic_infos[instr->intrinsic];
   assert(info->index_map[${enum}] > 0);
% if "struct" in data_type:
% if name == "io_semantics":
   val._pad = 0; /* clear padding bits */
% endif
   STATIC_ASSERT(sizeof(instr->const_index[0]) == sizeof(val));
   memcpy(&instr->const_index[info->index_map[${enum}] - 1], &val, sizeof(val));
% else:
   instr->const_index[info->index_map[${enum}] - 1] = val;
% endif
}

static inline bool
nir_intrinsic_has_${name}(const nir_intrinsic_instr *instr)
{
   const nir_intrinsic_info *info = &nir_intrinsic_infos[instr->intrinsic];
   return info->index_map[${enum}] > 0;
}
% endfor

#endif /* _NIR_INTRINSICS_INDICES_ */"""

from nir_intrinsics import INTR_INDICES
from mako.template import Template
import argparse
import os


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--outdir', required=True,
                        help='Directory to put the generated files in')

    args = parser.parse_args()

    path = os.path.join(args.outdir, 'nir_intrinsics_indices.h')
    with open(path, 'w') as f:
        f.write(Template(template).render(INTR_INDICES=INTR_INDICES))

if __name__ == '__main__':
    main()




#
# Copyright (c) 2020 Valve Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
import re
import sys
import os.path
import struct
import string
import copy
from math import floor

if os.isatty(sys.stdout.fileno()):
    set_red = "\033[31m"
    set_green = "\033[1;32m"
    set_normal = "\033[0m"
else:
    set_red = ''
    set_green = ''
    set_normal = ''

initial_code = '''
import re

def insert_code(code):
    insert_queue.append(CodeCheck(code, current_position))

def insert_pattern(pattern):
    insert_queue.append(PatternCheck(pattern, False, current_position))

def vector_gpr(prefix, name, size, align):
    insert_code(f'{name} = {name}0')
    for i in range(size):
        insert_code(f'{name}{i} = {name}0 + {i}')
    insert_code(f'success = {name}0 + {size - 1} == {name}{size - 1}')
    insert_code(f'success = {name}0 % {align} == 0')
    return f'{prefix}[#{name}0:#{name}{size - 1}]'

def sgpr_vector(name, size, align):
    return vector_gpr('s', name, size, align)

funcs.update({
    's64': lambda name: vector_gpr('s', name, 2, 2),
    's96': lambda name: vector_gpr('s', name, 3, 2),
    's128': lambda name: vector_gpr('s', name, 4, 4),
    's256': lambda name: vector_gpr('s', name, 8, 4),
    's512': lambda name: vector_gpr('s', name, 16, 4),
})
for i in range(2, 14):
    funcs['v%d' % (i * 32)] = lambda name: vector_gpr('v', name, i, 1)

def _match_func(names):
    for name in names.split(' '):
        insert_code(f'funcs["{name}"] = lambda _: {name}')
    return ' '.join(f'${name}' for name in names.split(' '))

funcs['match_func'] = _match_func

def search_re(pattern):
    global success
    success = re.search(pattern, output.read_line()) != None and success

'''

class Check:
    def __init__(self, data, position):
        self.data = data.rstrip()
        self.position = position

    def run(self, state):
        pass

class CodeCheck(Check):
    def run(self, state):
        indent = 0
        first_line = [l for l in self.data.split('\n') if l.strip() != ''][0]
        indent_amount = len(first_line) - len(first_line.lstrip())
        indent = first_line[:indent_amount]
        new_lines = []
        for line in self.data.split('\n'):
            if line.strip() == '':
                new_lines.append('')
                continue
            if line[:indent_amount] != indent:
                state.result.log += 'unexpected indent in code check:\n'
                state.result.log += self.data + '\n'
                return False
            new_lines.append(line[indent_amount:])
        code = '\n'.join(new_lines)

        try:
            exec(code, state.g)
            state.result.log += state.g['log']
            state.g['log'] = ''
        except BaseException as e:
            state.result.log += 'code check at %s raised exception:\n' % self.position
            state.result.log += code + '\n'
            state.result.log += str(e)
            return False
        if not state.g['success']:
            state.result.log += 'code check at %s failed:\n' % self.position
            state.result.log += code + '\n'
            return False
        return True

class StringStream:
    class Pos:
        def __init__(self):
            self.line = 1
            self.column = 1

    def __init__(self, data, name):
        self.name = name
        self.data = data
        self.offset = 0
        self.pos = StringStream.Pos()

    def reset(self):
        self.offset = 0
        self.pos = StringStream.Pos()

    def peek(self, num=1):
        return self.data[self.offset:self.offset+num]

    def peek_test(self, chars):
        c = self.peek(1)
        return c != '' and c in chars

    def read(self, num=4294967296):
        res = self.peek(num)
        self.offset += len(res)
        for c in res:
            if c == '\n':
                self.pos.line += 1
                self.pos.column = 1
            else:
                self.pos.column += 1
        return res

    def get_line(self, num):
        return self.data.split('\n')[num - 1].rstrip()

    def read_line(self):
        line = ''
        while self.peek(1) not in ['\n', '']:
            line += self.read(1)
        self.read(1)
        return line

    def skip_whitespace(self, inc_line):
        chars = [' ', '\t'] + (['\n'] if inc_line else [])
        while self.peek(1) in chars:
            self.read(1)

    def get_number(self):
        num = ''
        while self.peek() in string.digits:
            num += self.read(1)
        return num

    def check_identifier(self):
        return self.peek_test(string.ascii_letters + '_')

    def get_identifier(self):
        res = ''
        if self.check_identifier():
            while self.peek_test(string.ascii_letters + string.digits + '_'):
                res += self.read(1)
        return res

def format_error_lines(at, line_num, column_num, ctx, line):
    pred = '%s line %d, column %d of %s: "' % (at, line_num, column_num, ctx)
    return [pred + line + '"',
            '-' * (column_num - 1 + len(pred)) + '^']

class MatchResult:
    def __init__(self, pattern):
        self.success = True
        self.func_res = None
        self.pattern = pattern
        self.pattern_pos = StringStream.Pos()
        self.output_pos = StringStream.Pos()
        self.fail_message = ''

    def set_pos(self, pattern, output):
        self.pattern_pos.line = pattern.pos.line
        self.pattern_pos.column = pattern.pos.column
        self.output_pos.line = output.pos.line
        self.output_pos.column = output.pos.column

    def fail(self, msg):
        self.success = False
        self.fail_message = msg

    def format_pattern_pos(self):
        pat_pos = self.pattern_pos
        pat_line = self.pattern.get_line(pat_pos.line)
        res = format_error_lines('at', pat_pos.line, pat_pos.column, 'pattern', pat_line)
        func_res = self.func_res
        while func_res:
            pat_pos = func_res.pattern_pos
            pat_line = func_res.pattern.get_line(pat_pos.line)
            res += format_error_lines('in', pat_pos.line, pat_pos.column, func_res.pattern.name, pat_line)
            func_res = func_res.func_res
        return '\n'.join(res)

def do_match(g, pattern, output, skip_lines, in_func=False):
    assert(not in_func or not skip_lines)

    if not in_func:
        output.skip_whitespace(False)
    pattern.skip_whitespace(False)

    old_g = copy.copy(g)
    old_g_keys = list(g.keys())
    res = MatchResult(pattern)
    escape = False
    while True:
        res.set_pos(pattern, output)

        c = pattern.read(1)
        fail = False
        if c == '':
            break
        elif output.peek() == '':
            res.fail('unexpected end of output')
        elif c == '\\':
            escape = True
            continue
        elif c == '\n':
            old_line = output.pos.line
            output.skip_whitespace(True)
            if output.pos.line == old_line:
                res.fail('expected newline in output')
        elif not escape and c == '#':
            num = output.get_number()
            if num == '':
                res.fail('expected number in output')
            elif pattern.check_identifier():
                name = pattern.get_identifier()
                if name in g and int(num) != g[name]:
                    res.fail('unexpected number for \'%s\': %d (expected %d)' % (name, int(num), g[name]))
                elif name != '_':
                    g[name] = int(num)
        elif not escape and c == '$':
            name = pattern.get_identifier()

            val = ''
            while not output.peek_test(string.whitespace):
                val += output.read(1)

            if name in g and val != g[name]:
                res.fail('unexpected value for \'%s\': \'%s\' (expected \'%s\')' % (name, val, g[name]))
            elif name != '_':
                g[name] = val
        elif not escape and c == '%' and pattern.check_identifier():
            if output.read(1) != '%':
                res.fail('expected \'%\' in output')
            else:
                num = output.get_number()
                if num == '':
                    res.fail('expected number in output')
                else:
                    name = pattern.get_identifier()
                    if name in g and int(num) != g[name]:
                        res.fail('unexpected number for \'%s\': %d (expected %d)' % (name, int(num), g[name]))
                    elif name != '_':
                        g[name] = int(num)
        elif not escape and c == '@' and pattern.check_identifier():
            name = pattern.get_identifier()
            args = ''
            if pattern.peek_test('('):
                pattern.read(1)
                while pattern.peek() not in ['', ')']:
                    args += pattern.read(1)
                assert(pattern.read(1) == ')')
            func_res = g['funcs'][name](args)
            match_res = do_match(g, StringStream(func_res, 'expansion of "%s(%s)"' % (name, args)), output, False, True)
            if not match_res.success:
                res.func_res = match_res
                res.output_pos = match_res.output_pos
                res.fail(match_res.fail_message)
        elif not escape and c == ' ':
            while pattern.peek_test(' '):
                pattern.read(1)

            read_whitespace = False
            while output.peek_test(' \t'):
                output.read(1)
                read_whitespace = True
            if not read_whitespace:
                res.fail('expected whitespace in output, got %r' % (output.peek(1)))
        else:
            outc = output.peek(1)
            if outc != c:
                res.fail('expected %r in output, got %r' % (c, outc))
            else:
                output.read(1)
        if not res.success:
            if skip_lines and output.peek() != '':
                g.clear()
                g.update(old_g)
                res.success = True
                output.read_line()
                pattern.reset()
                output.skip_whitespace(False)
                pattern.skip_whitespace(False)
            else:
                return res

        escape = False

    if not in_func:
        while output.peek() in [' ', '\t']:
            output.read(1)

        if output.read(1) not in ['', '\n']:
            res.fail('expected end of output')
            return res

    return res

class PatternCheck(Check):
    def __init__(self, data, search, position):
        Check.__init__(self, data, position)
        self.search = search

    def run(self, state):
        pattern_stream = StringStream(self.data.rstrip(), 'pattern')
        res = do_match(state.g, pattern_stream, state.g['output'], self.search)
        if not res.success:
            state.result.log += 'pattern at %s failed: %s\n' % (self.position, res.fail_message)
            state.result.log += res.format_pattern_pos() + '\n\n'
            if not self.search:
                out_line = state.g['output'].get_line(res.output_pos.line)
                state.result.log += '\n'.join(format_error_lines('at', res.output_pos.line, res.output_pos.column, 'output', out_line))
            else:
                state.result.log += 'output was:\n'
                state.result.log += state.g['output'].data.rstrip() + '\n'
            return False
        return True

class CheckState:
    def __init__(self, result, variant, checks, output):
        self.result = result
        self.variant = variant
        self.checks = checks

        self.checks.insert(0, CodeCheck(initial_code, None))
        self.insert_queue = []

        self.g = {'success': True, 'funcs': {}, 'insert_queue': self.insert_queue,
                  'variant': variant, 'log': '', 'output': StringStream(output, 'output'),
                  'CodeCheck': CodeCheck, 'PatternCheck': PatternCheck,
                  'current_position': ''}

class TestResult:
    def __init__(self, expected):
        self.result = ''
        self.expected = expected
        self.log = ''

def check_output(result, variant, checks, output):
    state = CheckState(result, variant, checks, output)

    while len(state.checks):
        check = state.checks.pop(0)
        state.current_position = check.position
        if not check.run(state):
            result.result = 'failed'
            return

        for check in state.insert_queue[::-1]:
            state.checks.insert(0, check)
        state.insert_queue.clear()

    result.result = 'passed'
    return

def parse_check(variant, line, checks, pos):
    if line.startswith(';'):
        line = line[1:]
        if len(checks) and isinstance(checks[-1], CodeCheck):
            checks[-1].data += '\n' + line
        else:
            checks.append(CodeCheck(line, pos))
    elif line.startswith('!'):
        checks.append(PatternCheck(line[1:], False, pos))
    elif line.startswith('>>'):
        checks.append(PatternCheck(line[2:], True, pos))
    elif line.startswith('~'):
        end = len(line)
        start = len(line)
        for c in [';', '!', '>>']:
            if line.find(c) != -1 and line.find(c) < end:
                end = line.find(c)
        if end != len(line):
            match = re.match(line[1:end], variant)
            if match and match.end() == len(variant):
                parse_check(variant, line[end:], checks, pos)

def parse_test_source(test_name, variant, fname):
    in_test = False
    test = []
    expected_result = 'passed'
    line_num = 1
    for line in open(fname, 'r').readlines():
        if line.startswith('BEGIN_TEST(%s)' % test_name):
            in_test = True
        elif line.startswith('BEGIN_TEST_TODO(%s)' % test_name):
            in_test = True
            expected_result = 'todo'
        elif line.startswith('BEGIN_TEST_FAIL(%s)' % test_name):
            in_test = True
            expected_result = 'failed'
        elif line.startswith('END_TEST'):
            in_test = False
        elif in_test:
            test.append((line_num, line.strip()))
        line_num += 1

    checks = []
    for line_num, check in [(line_num, l[2:]) for line_num, l in test if l.startswith('//')]:
         parse_check(variant, check, checks, 'line %d of %s' % (line_num, os.path.split(fname)[1]))

    return checks, expected_result

def parse_and_check_test(test_name, variant, test_file, output, current_result):
    checks, expected = parse_test_source(test_name, variant, test_file)

    result = TestResult(expected)
    if len(checks) == 0:
        result.result = 'empty'
        result.log = 'no checks found'
    elif current_result != None:
        result.result, result.log = current_result
    else:
        check_output(result, variant, checks, output)
        if result.result == 'failed' and expected == 'todo':
            result.result = 'todo'

    return result

def print_results(results, output, expected):
    results = {name: result for name, result in results.items() if result.result == output}
    results = {name: result for name, result in results.items() if (result.result == result.expected) == expected}

    if not results:
        return 0

    print('%s tests (%s):' % (output, 'expected' if expected else 'unexpected'))
    for test, result in results.items():
        color = '' if expected else set_red
        print('   %s%s%s' % (color, test, set_normal))
        if result.log.strip() != '':
            for line in result.log.rstrip().split('\n'):
                print('      ' + line.rstrip())
    print('')

    return len(results)

def get_cstr(fp):
    res = b''
    while True:
        c = fp.read(1)
        if c == b'\x00':
            return res.decode('utf-8')
        else:
            res += c

if __name__ == "__main__":
   results = {}

   stdin = sys.stdin.buffer
   while True:
       packet_type = stdin.read(4)
       if packet_type == b'':
           break;

       test_name = get_cstr(stdin)
       test_variant = get_cstr(stdin)
       if test_variant != '':
           full_name = test_name + '/' + test_variant
       else:
           full_name = test_name

       test_source_file = get_cstr(stdin)
       current_result = None
       if ord(stdin.read(1)):
           current_result = (get_cstr(stdin), get_cstr(stdin))
       code_size = struct.unpack("=L", stdin.read(4))[0]
       code = stdin.read(code_size).decode('utf-8')

       results[full_name] = parse_and_check_test(test_name, test_variant, test_source_file, code, current_result)

   result_types = ['passed', 'failed', 'todo', 'empty']
   num_expected = 0
   num_unexpected = 0
   for t in result_types:
       num_expected += print_results(results, t, True)
   for t in result_types:
       num_unexpected += print_results(results, t, False)
   num_expected_skipped = print_results(results, 'skipped', True)
   num_unexpected_skipped = print_results(results, 'skipped', False)

   num_unskipped = len(results) - num_expected_skipped - num_unexpected_skipped
   color = set_red if num_unexpected else set_green
   print('%s%d (%.0f%%) of %d unskipped tests had an expected result%s' % (color, num_expected, floor(num_expected / num_unskipped * 100), num_unskipped, set_normal))
   if num_unexpected_skipped:
       print('%s%d tests had been unexpectedly skipped%s' % (set_red, num_unexpected_skipped, set_normal))

   if num_unexpected:
       sys.exit(1)




# (C) Copyright IBM Corporation 2004, 2005
# (C) Copyright Apple Inc. 2011
# Copyright (C) 2015 Intel Corporation
# All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# on the rights to use, copy, modify, merge, publish, distribute, sub
# license, and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
# IBM AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# Authors:
#    Jeremy Huddleston <jeremyhu@apple.com>
#
# Based on code ogiginally by:
#    Ian Romanick <idr@us.ibm.com>

import argparse

import license
import gl_XML, glX_XML

header = """/* GLXEXT is the define used in the xserver when the GLX extension is being
 * built.  Hijack this to determine whether this file is being built for the
 * server or the client.
 */
#ifdef HAVE_DIX_CONFIG_H
#include <dix-config.h>
#endif

#ifndef _WIN32
#include <dlfcn.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "main/glheader.h"

#include "glapi.h"
#include "glapitable.h"

#ifdef GLXEXT
#include "os.h"
#endif

static void
__glapi_gentable_NoOp(void) {
#if defined(GLXEXT)
    LogMessage(X_ERROR, "GLX: Call to unimplemented API: Unknown\\n");
#else
    fprintf(stderr, "Call to unimplemented API: Unknown\\n");
#endif
}

static void
__glapi_gentable_set_remaining_noop(struct _glapi_table *disp) {
    GLuint entries = _glapi_get_dispatch_table_size();
    void **dispatch = (void **) disp;
    unsigned i;

    /* ISO C is annoying sometimes */
    union {_glapi_proc p; void *v;} p;
    p.p = __glapi_gentable_NoOp;

    for(i=0; i < entries; i++)
        if(dispatch[i] == NULL)
            dispatch[i] = p.v;
}

"""

footer = """
struct _glapi_table *
_glapi_create_table_from_handle(void *handle, const char *symbol_prefix) {
    struct _glapi_table *disp = calloc(_glapi_get_dispatch_table_size(), sizeof(_glapi_proc));
    char symboln[512];

    if(!disp)
        return NULL;

    if(symbol_prefix == NULL)
        symbol_prefix = "";

    /* Note: This code relies on _glapi_table_func_names being sorted by the
     * entry point index of each function.
     */
    for (int func_index = 0; func_index < GLAPI_TABLE_COUNT; ++func_index) {
        const char *name = _glapi_table_func_names[func_index];
        void ** procp = &((void **)disp)[func_index];

        snprintf(symboln, sizeof(symboln), \"%s%s\", symbol_prefix, name);
#ifdef _WIN32
        *procp = GetProcAddress(handle, symboln);
#else
        *procp = dlsym(handle, symboln);
#endif
    }
    __glapi_gentable_set_remaining_noop(disp);

    return disp;
}

void
 _glapi_table_patch(struct _glapi_table *table, const char *name, void *wrapper)
{
   for (int func_index = 0; func_index < GLAPI_TABLE_COUNT; ++func_index) {
      if (!strcmp(_glapi_table_func_names[func_index], name)) {
            ((void **)table)[func_index] = wrapper;
            return;
         }
   }
   fprintf(stderr, "could not patch %s in dispatch table\\n", name);
}

"""


class PrintCode(gl_XML.gl_print_base):

    def __init__(self):
        gl_XML.gl_print_base.__init__(self)

        self.name = "gl_gentable.py (from Mesa)"
        self.license = license.bsd_license_template % ( \
"""Copyright (C) 1999-2001  Brian Paul   All Rights Reserved.
(C) Copyright IBM Corporation 2004, 2005
(C) Copyright Apple Inc 2011""", "BRIAN PAUL, IBM")

        return


    def get_stack_size(self, f):
        size = 0
        for p in f.parameterIterator():
            if p.is_padding:
                continue

            size += p.get_stack_size()

        return size


    def printRealHeader(self):
        print(header)
        return


    def printRealFooter(self):
        print(footer)
        return


    def printBody(self, api):

        # Determine how many functions have a defined offset.
        func_count = 0
        for f in api.functions_by_name.values():
            if f.offset != -1:
                func_count += 1

        # Build the mapping from offset to function name.
        funcnames = [None] * func_count
        for f in api.functions_by_name.values():
            if f.offset != -1:
                if not (funcnames[f.offset] is None):
                    raise Exception("Function table has more than one function with same offset (offset %d, func %s)" % (f.offset, f.name))
                funcnames[f.offset] = f.name

        # Check that the table has no gaps.  We expect a function at every offset,
        # and the code which generates the table relies on this.
        for i in range(0, func_count):
            if funcnames[i] is None:
                raise Exception("Function table has no function at offset %d" % (i))

        print("#define GLAPI_TABLE_COUNT %d" % func_count)
        print("static const char * const _glapi_table_func_names[GLAPI_TABLE_COUNT] = {")
        for i in range(0, func_count):
            print("    /* %5d */ \"%s\"," % (i, funcnames[i]))
        print("};")

        return


def _parser():
    """Parse arguments and return a namespace object."""
    parser = argparse.ArgumentParser()
    parser.add_argument('-f',
                        dest='filename',
                        default='gl_API.xml',
                        help='An XML file description of an API')

    return parser.parse_args()


def main():
    """Main function."""
    args = _parser()

    printer = PrintCode()

    api = gl_XML.parse_GL_API(args.filename, glX_XML.glx_item_factory())
    printer.Print(api)


if __name__ == '__main__':
    main()



# Copyright (C) 2012 Intel Corporation
# Copyright (C) 2021 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

# This script generates the file api_beginend_init.h.

import argparse
import collections
import license
import gl_XML
import sys
import apiexec
import re


class PrintCode(gl_XML.gl_print_base):
    def __init__(self):
        super().__init__()

        self.name = 'api_beginend_init_h.py'
        self.license = license.bsd_license_template % (
            'Copyright (C) 2012 Intel Corporation\n'
            'Copyright (C) 2021 Advanced Micro Devices, Inc.',
            'AUTHORS')

    def printBody(self, api):
        # Collect SET_* calls by the condition under which they should
        # be called.
        settings_by_condition = collections.defaultdict(lambda: [])

        for f in api.functionIterateAll():
            if f.exec_flavor != 'beginend':
                continue

            condition = apiexec.get_api_condition(f)
            if not condition:
                continue

            if (condition == '_mesa_is_desktop_gl(ctx) || ctx->API == API_OPENGLES2' and
                re.match('VertexAttrib[1-4].*ARB', f.name)):
                # These functions should map to an *ES callback for GLES2.
                settings_by_condition['_mesa_is_desktop_gl(ctx)'].append(
                    'SET_{0}(tab, NAME({0}));'.format(f.name))
                settings_by_condition['ctx->API == API_OPENGLES2'].append(
                    'SET_{0}(tab, NAME_ES({0}));'.format(f.name))
            else:
                macro = ('NAME_CALLLIST' if f.name[0:8] == 'CallList' else
                         'NAME_AE' if f.name == 'ArrayElement' else 'NAME')
                settings_by_condition[condition].append(
                    'SET_{0}(tab, {1}({0}));'.format(f.name, macro))

        # Print out an if statement for each unique condition, with
        # the SET_* calls nested inside it.
        for condition in sorted(settings_by_condition.keys()):
            print('if ({0}) {{'.format(condition))
            for setting in sorted(settings_by_condition[condition]):
                print('   {0}'.format(setting))
            print('}')


if __name__ == '__main__':
    apiexec.print_glapi_file(PrintCode())



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

# Copyright © 2021 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Helper script for manipulating the release calendar."""

from __future__ import annotations
import argparse
import csv
import contextlib
import datetime
import pathlib
import subprocess
import typing

if typing.TYPE_CHECKING:
    import _csv
    from typing_extensions import Protocol

    class RCArguments(Protocol):
        """Typing information for release-candidate command arguments."""

        manager: str

    class FinalArguments(Protocol):
        """Typing information for release command arguments."""

        series: str
        manager: str
        zero_released: bool

    class ExtendArguments(Protocol):
        """Typing information for extend command arguments."""

        series: str
        count: int


    CalendarRowType = typing.Tuple[typing.Optional[str], str, str, str, typing.Optional[str]]


_ROOT = pathlib.Path(__file__).parent.parent
CALENDAR_CSV = _ROOT / 'docs' / 'release-calendar.csv'
VERSION = _ROOT / 'VERSION'
LAST_RELEASE = 'This is the last planned release of the {}.x series.'
OR_FINAL = 'Or {}.0 final.'


def read_calendar() -> typing.List[CalendarRowType]:
    """Read the calendar and return a list of it's rows."""
    with CALENDAR_CSV.open('r') as f:
        return [typing.cast('CalendarRowType', tuple(r)) for r in csv.reader(f)]


def commit(message: str) -> None:
    """Commit the changes the the release-calendar.csv file."""
    subprocess.run(['git', 'commit', str(CALENDAR_CSV), '--message', message])



def _calculate_release_start(major: str, minor: str) -> datetime.date:
    """Calclulate the start of the release for release candidates.

    This is quarterly, on the second wednesday, in Januray, April, July, and Octobor.
    """
    quarter = datetime.date.fromisoformat(f'20{major}-0{[1, 4, 7, 10][int(minor)]}-01')

    # Wednesday is 3
    day = quarter.isoweekday()
    if day > 3:
        # this will walk back into the previous month, it's much simpler to
        # duplicate the 14 than handle the calculations for the month and year
        # changing.
        return quarter.replace(day=quarter.day - day + 3 + 14)
    elif day < 3:
        quarter = quarter.replace(day=quarter.day + 3 - day)
    return quarter.replace(day=quarter.day + 14)


def release_candidate(args: RCArguments) -> None:
    """Add release candidate entries."""
    with VERSION.open('r') as f:
        version = f.read().rstrip('-devel')
    major, minor, _ = version.split('.')
    date = _calculate_release_start(major, minor)

    data = read_calendar()

    with CALENDAR_CSV.open('w', newline='') as f:
        writer = csv.writer(f)
        writer.writerows(data)

        writer.writerow([f'{major}.{minor}', date.isoformat(), f'{major}.{minor}.0-rc1', args.manager])
        for row in range(2, 4):
            date = date + datetime.timedelta(days=7)
            writer.writerow([None, date.isoformat(), f'{major}.{minor}.0-rc{row}', args.manager])
        date = date + datetime.timedelta(days=7)
        writer.writerow([None, date.isoformat(), f'{major}.{minor}.0-rc4', args.manager, OR_FINAL.format(f'{major}.{minor}')])

    commit(f'docs: Add calendar entries for {major}.{minor} release candidates.')


def _calculate_next_release_date(next_is_zero: bool) -> datetime.date:
    """Calculate the date of the next release.

    If the next is .0, we have the release in seven days, if the next is .1,
    then it's in 14
    """
    date = datetime.date.today()
    day = date.isoweekday()
    if day < 3:
        delta = 3 - day
    elif day > 3:
        # this will walk back into the previous month, it's much simpler to
        # duplicate the 14 than handle the calculations for the month and year
        # changing.
        delta = (3 - day)
    else:
        delta = 0
    delta += 7
    if not next_is_zero:
        delta += 7
    return date + datetime.timedelta(days=delta)


def final_release(args: FinalArguments) -> None:
    """Add final release entries."""
    data = read_calendar()
    date = _calculate_next_release_date(not args.zero_released)

    with CALENDAR_CSV.open('w', newline='') as f:
        writer = csv.writer(f)
        writer.writerows(data)

        base = 1 if args.zero_released else 0

        writer.writerow([args.series, date.isoformat(), f'{args.series}.{base}', args.manager])
        for row in range(base + 1, 3):
            date = date + datetime.timedelta(days=14)
            writer.writerow([None, date.isoformat(), f'{args.series}.{row}', args.manager])
        date = date + datetime.timedelta(days=14)
        writer.writerow([None, date.isoformat(), f'{args.series}.3', args.manager, LAST_RELEASE.format(args.series)])

    commit(f'docs: Add calendar entries for {args.series} release.')


def extend(args: ExtendArguments) -> None:
    """Extend a release."""
    @contextlib.contextmanager
    def write_existing(writer: _csv._writer, current: typing.List[CalendarRowType]) -> typing.Iterator[CalendarRowType]:
        """Write the orinal file, yield to insert new entries.

        This is a bit clever, basically what happens it writes out the
        original csv file until it reaches the start of the release after the
        one we're appending, then it yields the last row. When control is
        returned it writes out the rest of the original calendar data.
        """
        last_row: typing.Optional[CalendarRowType] = None
        in_wanted = False
        for row in current:
            if in_wanted and row[0]:
                in_wanted = False
                assert last_row is not None
                yield last_row
            if row[0] == args.series:
                in_wanted = True
            if in_wanted and len(row) >= 5 and row[4] in {LAST_RELEASE.format(args.series), OR_FINAL.format(args.series)}:
                # If this was the last planned release and we're adding more,
                # then we need to remove that message and add it elsewhere
                r = list(row)
                r[4] = None
                # Mypy can't figure this out…
                row = typing.cast('CalendarRowType', tuple(r))
            last_row = row
            writer.writerow(row)
        # If this is the only entry we can hit a case where the contextmanager
        # hasn't yielded
        if in_wanted:
            yield row

    current = read_calendar()

    with CALENDAR_CSV.open('w', newline='') as f:
        writer = csv.writer(f)
        with write_existing(writer, current) as row:
            # Get rid of -rcX as well
            if '-rc' in row[2]:
                first_point = int(row[2].split('rc')[-1]) + 1
                template = '{}.0-rc{}'
                days = 7
            else:
                first_point = int(row[2].split('-')[0].split('.')[-1]) + 1
                template = '{}.{}'
                days = 14

            date = datetime.date.fromisoformat(row[1])
            for i in range(first_point, first_point + args.count):
                date = date + datetime.timedelta(days=days)
                r = [None, date.isoformat(), template.format(args.series, i), row[3], None]
                if i == first_point + args.count - 1:
                    if days == 14:
                        r[4] = LAST_RELEASE.format(args.series)
                    else:
                        r[4] = OR_FINAL.format(args.series)
                writer.writerow(r)

    commit(f'docs: Extend calendar entries for {args.series} by {args.count} releases.')


def main() -> None:
    parser = argparse.ArgumentParser()
    sub = parser.add_subparsers()

    rc = sub.add_parser('release-candidate', aliases=['rc'], help='Generate calendar entries for a release candidate.')
    rc.add_argument('manager', help="the name of the person managing the release.")
    rc.set_defaults(func=release_candidate)

    fr = sub.add_parser('release', help='Generate calendar entries for a final release.')
    fr.add_argument('manager', help="the name of the person managing the release.")
    fr.add_argument('series', help='The series to extend, such as "29.3" or "30.0".')
    fr.add_argument('--zero-released', action='store_true', help='The .0 release was today, the next release is .1')
    fr.set_defaults(func=final_release)

    ex = sub.add_parser('extend', help='Generate additional entries for a release.')
    ex.add_argument('series', help='The series to extend, such as "29.3" or "30.0".')
    ex.add_argument('count', type=int, help='The number of new entries to add.')
    ex.set_defaults(func=extend)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import sys
import serial

mode = sys.argv[1]
relay = sys.argv[2]

# our relays are "off" means "board is powered".
mode_swap = {
    "on": "off",
    "off": "on",
}
mode = mode_swap[mode]

ser = serial.Serial('/dev/ttyACM0', 115200, timeout=2)
command = "relay {} {}\n\r".format(mode, relay)
ser.write(command.encode())
ser.close()



#!/usr/bin/env python3
#
# Copyright © 2020 Google LLC
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import argparse
import subprocess
import re
from serial_buffer import SerialBuffer
import sys
import threading


class FastbootRun:
    def __init__(self, args, test_timeout):
        self.powerup = args.powerup
        self.ser = SerialBuffer(
            args.dev, "results/serial-output.txt", "R SERIAL> ")
        self.fastboot = "fastboot boot -s {ser} artifacts/fastboot.img".format(
            ser=args.fbserial)
        self.test_timeout = test_timeout

    def close(self):
        self.ser.close()

    def print_error(self, message):
        RED = '\033[0;31m'
        NO_COLOR = '\033[0m'
        print(RED + message + NO_COLOR)

    def logged_system(self, cmd, timeout=60):
        print("Running '{}'".format(cmd))
        try:
            return subprocess.call(cmd, shell=True, timeout=timeout)
        except subprocess.TimeoutExpired:
            self.print_error("timeout, restarting run...")
            return 2

    def run(self):
        if ret := self.logged_system(self.powerup):
            return ret

        fastboot_ready = False
        for line in self.ser.lines(timeout=2 * 60, phase="bootloader"):
            if re.search("fastboot: processing commands", line) or \
                    re.search("Listening for fastboot command on", line):
                fastboot_ready = True
                break

            if re.search("data abort", line):
                self.print_error(
                    "Detected crash during boot, restarting run...")
                return 2

        if not fastboot_ready:
            self.print_error(
                "Failed to get to fastboot prompt, restarting run...")
            return 2

        if ret := self.logged_system(self.fastboot):
            return ret

        print_more_lines = -1
        for line in self.ser.lines(timeout=self.test_timeout, phase="test"):
            if print_more_lines == 0:
                return 2
            if print_more_lines > 0:
                print_more_lines -= 1

            if re.search("---. end Kernel panic", line):
                return 1

            # The db820c boards intermittently reboot.  Just restart the run
            # when if we see a reboot after we got past fastboot.
            if re.search("PON REASON", line):
                self.print_error(
                    "Detected spontaneous reboot, restarting run...")
                return 2

            # db820c sometimes wedges around iommu fault recovery
            if re.search("watchdog: BUG: soft lockup - CPU.* stuck", line):
                self.print_error(
                    "Detected kernel soft lockup, restarting run...")
                return 2

            # If the network device dies, it's probably not graphics's fault, just try again.
            if re.search("NETDEV WATCHDOG", line):
                self.print_error(
                    "Detected network device failure, restarting run...")
                return 2

            # A3xx recovery doesn't quite work. Sometimes the GPU will get
            # wedged and recovery will fail (because power can't be reset?)
            # This assumes that the jobs are sufficiently well-tested that GPU
            # hangs aren't always triggered, so just try again. But print some
            # more lines first so that we get better information on the cause
            # of the hang. Once a hang happens, it's pretty chatty.
            if "[drm:adreno_recover] *ERROR* gpu hw init failed: -22" in line:
                self.print_error(
                    "Detected GPU hang, restarting run...")
                if print_more_lines == -1:
                    print_more_lines = 30

            result = re.search("hwci: mesa: (\S*)", line)
            if result:
                if result.group(1) == "pass":
                    return 0
                else:
                    return 1

        self.print_error(
            "Reached the end of the CPU serial log without finding a result, restarting run...")
        return 2


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dev', type=str, help='Serial device (otherwise reading from serial-output.txt)')
    parser.add_argument('--powerup', type=str,
                        help='shell command for rebooting', required=True)
    parser.add_argument('--powerdown', type=str,
                        help='shell command for powering off', required=True)
    parser.add_argument('--fbserial', type=str,
                        help='fastboot serial number of the board', required=True)
    parser.add_argument('--test-timeout', type=int,
                        help='Test phase timeout (minutes)', required=True)
    args = parser.parse_args()

    fastboot = FastbootRun(args, args.test_timeout * 60)

    while True:
        retval = fastboot.run()
        fastboot.close()
        if retval != 2:
            break

        fastboot = FastbootRun(args, args.test_timeout * 60)

    fastboot.logged_system(args.powerdown)

    sys.exit(retval)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import re
import sys
import gzip
import io

# Captures per-frame state, including all the renderpasses, and
# time spent in blits and compute jobs:
class Frame:
    def __init__(self):
        self.frame_nr = None
        self.renderpasses = []
        # Times in ns:
        self.times_sysmem = []
        self.times_gmem = []
        self.times_compute = []
        self.times_blit = []

    def print(self):
        print("FRAME[{}]: {} blits ({:,} ns), {} SYSMEM ({:,} ns), {} GMEM ({:,} ns), {} COMPUTE ({:,} ns)".format(
                self.frame_nr,
                len(self.times_blit),    sum(self.times_blit),
                len(self.times_sysmem),  sum(self.times_sysmem),
                len(self.times_gmem),    sum(self.times_gmem),
                len(self.times_compute), sum(self.times_compute)
            ))

        i = 0
        prologue_time = 0
        binning_time = 0
        restore_clear_time = 0
        draw_time = 0
        resolve_time = 0
        elapsed_time = 0
        total_time = sum(self.times_blit) + sum(self.times_sysmem) + sum(self.times_gmem) + sum(self.times_compute)

        for renderpass in self.renderpasses:
            renderpass.print(i)
            prologue_time += renderpass.prologue_time
            binning_time += renderpass.binning_time
            restore_clear_time += renderpass.restore_clear_time
            draw_time += renderpass.draw_time
            resolve_time += renderpass.resolve_time
            elapsed_time += renderpass.elapsed_time
            i += 1

        print("  TOTAL: prologue: {:,} ns ({}%), binning: {:,} ns ({}%), restore/clear: {:,} ns ({}%), draw: {:,} ns ({}%), resolve: {:,} ns ({}%), blit: {:,} ns ({}%), compute: {:,} ns ({}%), GMEM: {:,} ns ({}%), sysmem: {:,} ns ({}%), total: {:,} ns\n".format(
                prologue_time, 100.0 * prologue_time / total_time,
                binning_time, 100.0 * binning_time / total_time,
                restore_clear_time, 100.0 * restore_clear_time / total_time,
                draw_time, 100.0 * draw_time / total_time,
                resolve_time, 100.0 * resolve_time / total_time,
                sum(self.times_blit), 100.0 * sum(self.times_blit) / total_time,
                sum(self.times_compute), 100.0 * sum(self.times_compute) / total_time,
                sum(self.times_gmem), 100.0 * sum(self.times_gmem) / total_time,
                sum(self.times_sysmem), 100.0 * sum(self.times_sysmem) / total_time,
                total_time
            ))

class FramebufferState:
    def __init__(self, width, height, layers, samples, nr_cbufs):
        self.width = width
        self.height = height
        self.layers = layers
        self.samples = samples
        self.nr_cbufs = nr_cbufs
        self.surfaces = []        # per MRT + zsbuf

    def get_formats(self):
        formats = []
        for surface in self.surfaces:
            formats.append(surface.format)
        return formats

class SurfaceState:
    def __init__(self, width, height, samples, format):
        self.width = width
        self.height = height
        self.samples = samples
        self.format = format

class BinningState:
    def __init__(self, nbins_x, nbins_y, bin_w, bin_h):
        self.nbins_x = nbins_x
        self.nbins_y = nbins_y
        self.bin_w = bin_w
        self.bin_h = bin_h

# Captures per-renderpass state, which can be either a binning or
# sysmem pass.  Blits and compute jobs are not tracked separately
# but have their time their times accounted for in the Frame state
class RenderPass:
    def __init__(self, cleared, gmem_reason, num_draws):
        self.cleared = cleared
        self.gmem_reason = gmem_reason
        self.num_draws = num_draws

        # The rest of the parameters aren't known until we see a later trace:
        self.binning_state = None   # None for sysmem passes, else BinningState
        self.fb = None
        self.fast_cleared = None

        self.elapsed_time = 0
        self.state_restore_time = 0
        self.prologue_time = 0
        self.draw_time = 0
        self.restore_clear_time = 0

        # Specific to GMEM passes:
        self.binning_time = 0
        self.vsc_overflow_test_time = 0
        self.resolve_time = 0

    def print_gmem_pass(self, nr):
        print("  GMEM[{}]: {}x{} ({}x{} tiles), {} draws, prologue: {:,} ns, binning: {:,} ns, restore/clear: {:,} ns, draw: {:,} ns, resolve: {:,} ns, total: {:,} ns, rt/zs: {}".format(
                nr, self.fb.width, self.fb.height,
                self.binning_state.nbins_x, self.binning_state.nbins_y,
                self.num_draws, self.prologue_time, self.binning_time,
                self.restore_clear_time, self.draw_time, self.resolve_time,
                self.elapsed_time,
                ", ".join(self.fb.get_formats())
            ))

    def print_sysmem_pass(self, nr):
        print("  SYSMEM[{}]: {}x{}, {} draws, prologue: {:,} ns, clear: {:,} ns, draw: {:,} ns, total: {:,} ns, rt/zs: {}".format(
                nr, self.fb.width, self.fb.height,
                self.num_draws, self.prologue_time,
                self.restore_clear_time, self.draw_time,
                self.elapsed_time,
                ", ".join(self.fb.get_formats())
            ))

    def print(self, nr):
        if self.binning_state:
            self.print_gmem_pass(nr)
        else:
            self.print_sysmem_pass(nr)

def main():
    filename = sys.argv[1]
    if filename.endswith(".gz"):
        file = gzip.open(filename, "r")
        file = io.TextIOWrapper(file)
    else:
        file = open(filename, "r")
    lines = file.read().split('\n')

    flush_batch_match = re.compile(r": flush_batch: (\S+): cleared=(\S+), gmem_reason=(\S+), num_draws=(\S+)")
    framebuffer_match = re.compile(r": framebuffer: (\S+)x(\S+)x(\S+)@(\S+), nr_cbufs: (\S+)")
    surface_match     = re.compile(r": surface: (\S+)x(\S+)@(\S+), fmt=(\S+)")

    # draw/renderpass passes:
    gmem_match          = re.compile(r": render_gmem: (\S+)x(\S+) bins of (\S+)x(\S+)")
    sysmem_match        = re.compile(r": render_sysmem")
    state_restore_match = re.compile(r"\+(\S+): end_state_restore")
    prologue_match      = re.compile(r"\+(\S+): end_prologue")
    binning_ib_match    = re.compile(r"\+(\S+): end_binning_ib")
    vsc_overflow_match  = re.compile(r"\+(\S+): end_vsc_overflow_test")
    draw_ib_match       = re.compile(r"\+(\S+): end_draw_ib")
    resolve_match       = re.compile(r"\+(\S+): end_resolve")
    
    start_clear_restore_match = re.compile(r"start_clear_restore: fast_cleared: (\S+)")
    end_clear_restore_match   = re.compile(r"\+(\S+): end_clear_restore")

    # Non-draw passes:
    compute_match = re.compile(r": start_compute")
    blit_match    = re.compile(r": start_blit")

    # End of pass/frame markers:
    elapsed_match = re.compile(r"ELAPSED: (\S+) ns")
    eof_match     = re.compile(r"END OF FRAME (\S+)")

    frame = Frame()      # current frame state
    renderpass = None    # current renderpass state
    times = None

    # Helper to set the appropriate times table for the current pass,
    # which is expected to only happen once for a given render pass
    def set_times(t):
        nonlocal times
        if times  is not None:
            print("expected times to not be set yet")
        times = t

    for line in lines:
        # Note, we only expect the flush_batch trace for !nondraw:
        match = re.search(flush_batch_match, line)
        if match is not None:
            assert(renderpass is None)
            renderpass = RenderPass(cleared=match.group(2),
                                    gmem_reason=match.group(3),
                                    num_draws=match.group(4))
            frame.renderpasses.append(renderpass)
            continue

        match = re.search(framebuffer_match, line)
        if match is not None:
            assert(renderpass.fb is None)
            renderpass.fb = FramebufferState(width=match.group(1),
                                             height=match.group(2),
                                             layers=match.group(3),
                                             samples=match.group(4),
                                             nr_cbufs=match.group(5))
            continue

        match = re.search(surface_match, line)
        if match is not None:
            surface = SurfaceState(width=match.group(1),
                                   height=match.group(2),
                                   samples=match.group(3),
                                   format=match.group(4))
            renderpass.fb.surfaces.append(surface)
            continue

        match = re.search(gmem_match, line)
        if match is not None:
            assert(renderpass.binning_state is None)
            renderpass.binning_state = BinningState(nbins_x=match.group(1),
                                                    nbins_y=match.group(2),
                                                    bin_w=match.group(3),
                                                    bin_h=match.group(4))
            set_times(frame.times_gmem)
            continue

        match = re.search(sysmem_match, line)
        if match is not None:
            assert(renderpass.binning_state is None)
            set_times(frame.times_sysmem)
            continue

        match = re.search(state_restore_match, line)
        if match is not None:
            renderpass.state_restore_time += int(match.group(1))
            continue

        match = re.search(prologue_match, line)
        if match is not None:
            renderpass.prologue_time += int(match.group(1))
            continue

        match = re.search(binning_ib_match, line)
        if match is not None:
            assert(renderpass.binning_state is not None)
            renderpass.binning_time += int(match.group(1))
            continue

        match = re.search(vsc_overflow_match, line)
        if match is not None:
            assert(renderpass.binning_state is not None)
            renderpass.vsc_overflow_test_time += int(match.group(1))
            continue

        match = re.search(draw_ib_match, line)
        if match is not None:
            renderpass.draw_time += int(match.group(1))
            continue

        match = re.search(resolve_match, line)
        if match is not None:
            assert(renderpass.binning_state is not None)
            renderpass.resolve_time += int(match.group(1))
            continue

        match = re.search(start_clear_restore_match, line)
        if match is not None:
            renderpass.fast_cleared = match.group(1)
            continue

        match = re.search(end_clear_restore_match, line)
        if match is not None:
            renderpass.restore_clear_time += int(match.group(1))
            continue

        match = re.search(compute_match, line)
        if match is not None:
            set_times(frame.times_compute)
            continue

        match = re.search(blit_match, line)
        if match is not None:
            set_times(frame.times_blit)
            continue

        match = re.search(eof_match, line)
        if match is not None:
            frame.frame_nr = int(match.group(1))
            frame.print()
            frame = Frame()
            times = None
            renderpass = None
            continue

        match = re.search(elapsed_match, line)
        if match is not None:
            time = int(match.group(1))
            #print("ELAPSED: " + str(time) + " ns")
            if renderpass is not None:
                renderpass.elapsed_time = time
            times.append(time)
            times = None
            renderpass = None
            continue


if __name__ == "__main__":
    main()

